// Copyright 2005-2011, FreeHEP.
package hep.wired.cut;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.util.*;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BoxLayout;
import javax.swing.DefaultCellEditor;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.RowFilter;
import javax.swing.RowFilter.Entry;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.plaf.UIResource;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableRowSorter;


/**
 * Class that maintains JTable of conditions in a {@link Cut} and controls that manipulate them.
 *
 * @author Dmitry Onoprienko
 * @version $Id: $
 */
public class CutEditor implements ChangeListener {

  private enum Column {
    APPLY(" Apply ", Boolean.class),
    NAME(" Name ", Condition.class),
    STATUS(" Accept if ", Condition.class),
    INVERT(" Invert ", Condition.class);
    public final Class klass;
    public final String header;
    private Column(String header, Class klass) {
      this.header = header;
      this.klass = klass;
    }
    public String toString() {return header;}
  };

// -- Private parts : ----------------------------------------------------------

  Cut _cut;

  protected JTable _table;
  protected ConditionTableModel _model;
  protected TableRowSorter<ConditionTableModel> _sorter;
  protected ListSelectionListener _tableSelectionListener;

  protected JCheckBox _hide;
  protected JButton _actions;
  protected Action _actNew;
  protected Action _actAdd;
  protected Action _actDelete;
  protected Action _actDeleteAll;
  protected Action _actDeleteDisabled;
  protected Action _actReset;
  protected Action _actResetAll;
  protected Action _actSync;

  protected EnumMap<ConditionDefinition.Type, ConditionEditor> _conditionEditors;
  protected ConditionEditor _currentEditor;
  protected CutManager _manager;
  protected int _editRow = -1;

  protected boolean _enabled;

// -- Construction and initialization : ----------------------------------------

  public CutEditor() {

    _conditionEditors = new EnumMap(ConditionDefinition.Type.class);

    // Create table of attributes and set tool tips

    _model = new ConditionTableModel();
    _table = new JTable(_model) {
      public String getToolTipText(MouseEvent e) {
        java.awt.Point p = e.getPoint();
        int row = convertRowIndexToModel(rowAtPoint(p));
        int col = convertColumnIndexToModel(columnAtPoint(p));
        Condition condition = _cut.getCondition(row);
        String tip = null;
        String extra = null;
        switch (Column.values()[col]) {
          case APPLY:
            tip = "Check box to enable condition";
            extra = condition.getDefinition().getDescription();
            break;
          case NAME:
            tip = condition.getDefinition().getName();
            extra = condition.getDefinition().getDescription();
            break;
          case STATUS:
            tip = condition.getStatus();
            extra = condition.getDefinition().getUnit();
            break;
          case INVERT:
            tip = "Check box to invert condition";
            extra = condition.getDefinition().getDescription();
            break;
          default:
            return super.getToolTipText(e);
        }
        if (extra != null) {
          tip = (new StringBuilder("<html><table><tr><td>")).append(tip).append("</td></tr><tr><td>")
                .append(extra).append("</td></tr></table></html>").toString();
        }
        return tip;
      }
      public void changeSelection(int row, int column, boolean toggle, boolean extend) {
        super.changeSelection(row, column, toggle, extend);
        row = convertRowIndexToModel(row);
        column = convertColumnIndexToModel(column);
        if (row == _editRow) {  // selection is the same condition as previously edited
          if (column == Column.APPLY.ordinal() || column == Column.NAME.ordinal()) {
            _editRow = -1;
            showConditionEditor(null);
          } else if (column == Column.INVERT.ordinal()) {
            showConditionEditor(_conditionEditors.get(_cut.getCondition(row).getDefinition().getType()));
          }
        } else { // selection is not the same condition as previously edited
          Condition condition = _cut.getCondition(row);
          ConditionEditor editor = _conditionEditors.get(condition.getDefinition().getType());
          if (editor == null || !condition.isEnabled() || column != Column.STATUS.ordinal()) {  // new edit not triggered
            if (_editRow != -1) {
              _editRow = -1;
              showConditionEditor(null);
            }
          } else {  // new edit triggered
            _editRow = row;
            editor.setCondition(condition);
            showConditionEditor(editor);
          }
        }
      }
    };

    // Table - selection and sorting modes

    _table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    _sorter = new TableRowSorter(_model);
    _sorter.setRowFilter(new RowFilter<ConditionTableModel, Integer>() {
      public boolean include(Entry<? extends ConditionTableModel, ? extends Integer> entry) {
        if (_hide.isSelected()) {
          Condition c = _cut.getCondition(entry.getIdentifier());
          return c.isEnabled();
        } else {
          return true;
        }
      }
    });
    _table.setRowSorter(_sorter);

    // Table - initial and minimum width for columns

    TableCellRenderer headerRenderer = _table.getTableHeader().getDefaultRenderer();
    for (Column col : Column.values()) {
      TableColumn column = _table.getColumnModel().getColumn(col.ordinal());
      int headerWidth = headerRenderer.getTableCellRendererComponent(null, column.getHeaderValue(),false, false, 0, 0).getPreferredSize().width;
      int cellWidth = col.klass == Boolean.class ?
          _table.getDefaultRenderer(col.klass).getTableCellRendererComponent(_table, true, true, true, 0, 0).getPreferredSize().width
          : 400;
      int width = Math.max(headerWidth, cellWidth);
      column.setPreferredWidth(width);
      if (col.klass == Boolean.class) column.setMinWidth(width);
    }

    // Table - custom renderers and editors

    TableColumn column = _table.getColumnModel().getColumn(Column.NAME.ordinal());
    column.setCellRenderer(new DefaultTableCellRenderer(){
      public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
        Condition c = (Condition)value;
        Component out = super.getTableCellRendererComponent(table, c.getDefinition().getName(), isSelected, hasFocus, row, column);
        out.setEnabled(c.isEnabled());
        return out;
      }
    });

    column = _table.getColumnModel().getColumn(Column.STATUS.ordinal());
    column.setCellRenderer(new DefaultTableCellRenderer(){
      public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
        Condition c = (Condition)value;
        Component out = super.getTableCellRendererComponent(table, c.getStatus(), isSelected, hasFocus, row, column);
        out.setEnabled(c.isEnabled());
        return out;
      }
    });

    column = _table.getColumnModel().getColumn(Column.INVERT.ordinal());
    column.setCellRenderer(new InvertRenderer());
    column.setCellEditor(new DefaultCellEditor(new JCheckBox()) {
      public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row, int column) {
        Component c = super.getTableCellEditorComponent(table, ((Condition)value).isInverted(), isSelected, row, column);
        c.setEnabled(((Condition)value).isEnabled());
        ((JCheckBox)c).setHorizontalAlignment(JLabel.CENTER);
        return c;
      }
    });

    // Hide disabled conditions checkbox

    _hide = new JCheckBox("Hide disabled");
    _hide.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        _sorter.allRowsChanged();
      }
    });

    // Actions button

    _actNew = new AbstractAction("New...") {
      public void actionPerformed(ActionEvent e) {
        System.out.println("Not supported yet.");
      }
    };
    _actNew.putValue(Action.SHORT_DESCRIPTION, "Create new condition");

    _actAdd = new AbstractAction("Add...") {
      public void actionPerformed(ActionEvent e) {
        System.out.println("Not supported yet.");
      }
    };
    _actAdd.putValue(Action.SHORT_DESCRIPTION, "Add condition based on one of the existing definitions");

    _actDelete = new AbstractAction("Delete") {
      public void actionPerformed(ActionEvent e) {
        if (_cut != null) _cut.remove(getSelectedCondition());
        _model.fireTableDataChanged();
      }
    };
    _actDelete.putValue(Action.SHORT_DESCRIPTION, "Delete selected condition");

    _actDeleteAll = new AbstractAction("Delete All") {
      public void actionPerformed(ActionEvent e) {
        if (_cut != null) _cut.removeAll();
        _model.fireTableDataChanged();
      }
    };
    _actDeleteAll.putValue(Action.SHORT_DESCRIPTION, "Delete all conditions");

    _actDeleteDisabled = new AbstractAction("Delete disabled") {
      public void actionPerformed(ActionEvent e) {
        if (_cut != null) _cut.removeDisabled();
        _model.fireTableDataChanged();
      }
    };
    _actDeleteDisabled.putValue(Action.SHORT_DESCRIPTION, "Delete disabled conditions");

    _actReset = new AbstractAction("Reset") {
      public void actionPerformed(ActionEvent e) {
        Condition c = getSelectedCondition();
        c.reset();
        if (_currentEditor == null)  _currentEditor.setCondition(c);
     }
    };
    _actReset.putValue(Action.SHORT_DESCRIPTION, "<html>Reset selected condition<p>"
            + "Resetting a condition makes it accept all values if the condition is not inverted, "
            + "reject all values if the condition is inverted.</html>");

    _actResetAll = new AbstractAction("Reset all") {
      public void actionPerformed(ActionEvent e) {
        if (_cut != null) {
          for (Condition c : _cut.getConditions()) c.reset();
        }
        if (_currentEditor == null)  _currentEditor.setCondition(_currentEditor.getCondition());
      }
    };
    _actResetAll.putValue(Action.SHORT_DESCRIPTION, "<html>Reset all conditions<p>"
            + "Resetting a condition makes it accept all values if the condition is not inverted, "
            + "reject all values if the condition is inverted.</html>");

    _actSync = new AbstractAction("Synchronize") {
      public void actionPerformed(ActionEvent e) {
        if (_cut != null) _cut.synchronize(_manager.getCurrentDefinitions(_cut));
        _model.fireTableDataChanged();
      }
    };
    _actSync.putValue(Action.SHORT_DESCRIPTION, "<html>Synchronize to current record<p>"
            + "Delete conditions if there are no objects with identically named attributes in the current record.<br>"
            + "Add missing conditions corresponding to attributes in the record.</html>");

    final JPopupMenu pm = new JPopupMenu();
//    pm.add(_actNew);
//    pm.add(_actAdd);
    pm.add(_actDelete);
    pm.add(_actDeleteAll);
    pm.add(_actDeleteDisabled);
    pm.add(_actSync);
    pm.addSeparator();
    pm.add(_actReset);
    pm.add(_actResetAll);
    _actions = new JButton("Actions...");
    _actions.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        pm.show(_actions, 0, _actions.getHeight());
      }
    });

    _tableSelectionListener = new ListSelectionListener() {
      public void valueChanged(ListSelectionEvent e) {
        if (e == null || !e.getValueIsAdjusting()) {
          Condition c = getSelectedCondition();
          if (c == null) {
            _actDelete.setEnabled(false);
            _actReset.setEnabled(false);
            boolean notempty = _cut != null && _cut.size() != 0;
            _actDeleteAll.setEnabled(notempty);
            _actDeleteDisabled.setEnabled(notempty);
            _actResetAll.setEnabled(notempty);
            showConditionEditor(null);
          } else {
            _actDelete.setEnabled(true);
            _actDeleteAll.setEnabled(true);
            _actDeleteDisabled.setEnabled(true);
            _actResetAll.setEnabled(true);
            _actReset.setEnabled(c.getDefinition().getType() != ConditionDefinition.Type.BOOLEAN);
          }
        }
      }
    };
    _table.getSelectionModel().addListSelectionListener(_tableSelectionListener);

    _enabled = false;
  }


// -- Setters : ----------------------------------------------------------------

  public void setCut(Cut cut) {
    if (_cut != null) _cut.removeChangeListener(this);
    _cut = cut;
    _editRow = -1;
    if (_cut != null) {
      _cut.addChangeListener(this);
      if (_table.getSelectionModel().isSelectionEmpty()) {
        _tableSelectionListener.valueChanged(null);
      } else {
        _table.clearSelection();
      }
    }
    _model.fireTableDataChanged();
  }

  public void setManager(CutManager manager) {
    _manager = manager;
  }

  public void setConditionEditor(ConditionDefinition.Type type, ConditionEditor editor) {
    _conditionEditors.put(type, editor);
  }

  public void setEnabled(boolean state) {
    _enabled = state;
    _hide.setEnabled(state);
    _actions.setEnabled(state);
  }


// -- Access to components : ---------------------------------------------------

  public JTable getConditionsTable() {
    return _table;
  }

  public JButton getActionsButton() {
    return _actions;
  }

  public JCheckBox getHideDisabledBox() {
    return _hide;
  }

  public ConditionEditor getConditionEditor(ConditionDefinition.Type conditionType) {
    return _conditionEditors.get(conditionType);
  }


// -- Listening to changes in the cut : ----------------------------------------


  public void stateChanged(ChangeEvent e) {
    try {
      Condition condition = (Condition) e.getSource();
      int row = _cut.getConditions().indexOf(condition);
      if (row != -1) {
        _model.fireTableRowsUpdated(row, row);
        _table.repaint();
      }
    } catch (ClassCastException x) {}
  }

// -- Conditions table model : -------------------------------------------------


  /** Model for a table of {@link Cut}'s conditions. */
  protected class ConditionTableModel extends AbstractTableModel {

    protected ConditionTableModel() {
    }

    public Class<?> getColumnClass(int column) {
      return Column.values()[column].klass;
    }

    public String getColumnName(int column) {
      return Column.values()[column].header;
    }

    public boolean isCellEditable(int row, int column) {
      if (!_enabled) return false;
      switch (Column.values()[column]) {
        case APPLY:
        case INVERT:
          return true;
        case NAME:
        case STATUS:
          return false;
        default:
          throw new IllegalArgumentException();
      }
    }

    public void setValueAt(Object aValue, int row, int column) {
      Condition condition = _cut.getCondition(row);
      switch (Column.values()[column]) {
        case APPLY:
          condition.setEnabled((Boolean)aValue);
          break;
        case INVERT:
          condition.setInverted((Boolean)aValue);
          break;
        default:
          return;
      }
      fireTableRowsUpdated(row, row);
    }

    public int getColumnCount() {
      return Column.values().length;
    }

    public int getRowCount() {
      return _cut == null ? 0 : _cut.size();
    }

    public Object getValueAt(int row, int column) {
      Condition condition = _cut.getCondition(row);
      switch (Column.values()[column]) {
        case APPLY:
          return condition.isEnabled();
        default:
          return condition;
      }
    }
  }


// -- INVERT renderer class : --------------------------------------------------

  private class InvertRenderer extends JCheckBox implements TableCellRenderer, UIResource {

    public InvertRenderer() {
      super();
      setHorizontalAlignment(JLabel.CENTER);
      setBorderPainted(true);
      setBorder(new EmptyBorder(1, 1, 1, 1));
    }

    public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
      try {
        Condition c = (Condition)value;
        if (isSelected) {
          setForeground(table.getSelectionForeground());
          setBackground(table.getSelectionBackground());
        } else {
          setForeground(table.getForeground());
          setBackground(table.getBackground());
        }
        setSelected(c.isInverted());
        setEnabled(c.isEnabled());
      } catch (ClassCastException x) {  // appease DefaultCellEditor trying to match border/background to renderer
        setSelected((Boolean)value);
        setForeground(table.getSelectionForeground());
        setBackground(table.getSelectionBackground());
      }
      return this;
    }

  }


// -- Actions on conditions : --------------------------------------------------

  /**
   * Returns currently selected condition.
   * Returns <tt>null</tt> if no condition is selected.
   */
  public Condition getSelectedCondition() {
    if (_cut == null) return null;
    int row = _table.getSelectedRow();
    if (row == -1) {
      return null;
    } else {
      return _cut.getCondition(_table.convertRowIndexToModel(row));
    }
  }


// -- Local methods : ----------------------------------------------------------

  protected void showConditionEditor(ConditionEditor editor) {
    _currentEditor = editor;
    if (_manager != null) _manager.showConditionEditor(editor);
  }


// -- Testing code : -----------------------------------------------------------

  public static void main(String[] args) {
    javax.swing.SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        createAndShowGUI();
      }
    });
  }

  private static void createAndShowGUI() {

    JFrame frame = new JFrame("TableTest");
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

    JPanel newContentPane = new JPanel();
    newContentPane.setOpaque(true);
    frame.setContentPane(newContentPane);

    doTest(newContentPane);

    frame.pack();
    frame.setVisible(true);
  }

  private static class TestDisplay extends JFrame implements CutManager {
    JPanel contentPane;

    public TestDisplay() {
      super("Display");
      contentPane = new JPanel();
      contentPane.setOpaque(true);
      setContentPane(contentPane);
      pack();
      setVisible(true);
    }

    public void showConditionEditor(ConditionEditor editor) {
      setVisible(false);
      contentPane.removeAll();
      if (editor != null) contentPane.add(editor.getPanel());
      pack();
      setVisible(true);
    }

    public Set<ConditionDefinition> getCurrentDefinitions() {
      throw new UnsupportedOperationException("Not supported yet.");
    }
    public Set<ConditionDefinition> getCurrentDefinitions(Cut cut) {
      throw new UnsupportedOperationException("Not supported yet.");
    }
    public Set<ConditionDefinition> getDefinitions() {
      throw new UnsupportedOperationException("Not supported yet.");
    }
    public ConditionDefinition getDefinition(String name) {
      throw new UnsupportedOperationException("Not supported yet.");
    }
  }

  private static void doTest(final JPanel panel) {

    Cut cut = new Cut("Test Cut");
    cut.add(new BooleanCondition(new BooleanConditionDefinition("Boolean Condition 1", "Description of boolean condition 1")));
    cut.add(new BooleanCondition(new BooleanConditionDefinition("Boolean Condition 2", "Description of boolean condition 2")));
    cut.add(new BooleanCondition(new BooleanConditionDefinition("Boolean Condition 3")));
    StringCondition sc1 = new StringCondition(new StringConditionDefinition("String Condition 1"));
    sc1.set(Arrays.asList(new String[]{"string 1", "string 3", "string 2"}));
    cut.add(sc1);

    CutEditor cutEditor = new CutEditor();
    cutEditor.setCut(cut);

    panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
    panel.add(new JScrollPane(cutEditor.getConditionsTable()));

    TestDisplay display = new TestDisplay();
    cutEditor.setManager(display);
  }

// -----------------------------------------------------------------------------
}
