package hep.wired.cut;

import java.util.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import hep.wired.util.SwingEventMulticaster;
import org.freehep.util.Value;

/**
 * Set of cuts associated with an event display view.
 *
 * @author Dmitry Onoprienko
 */
public class CutSet implements ChangeListener {

// -- Private parts : ----------------------------------------------------------

  private Cut[] _cuts; // enabled cuts
  private boolean[] _cutInverted; // inversion states for enabled cuts
  private Map<Cut, Boolean> _cutMap; // map of all cuts to their inversion states

  protected boolean _enabled;
  protected boolean _inverted;
  protected boolean _passThrough;

  private ChangeListener _listener;
  protected ChangeEvent _defaultEvent;

  private transient Set<String> _conditionNames; // names of enabled conditions in all cuts

// -- Construction : -----------------------------------------------------------

  public CutSet() {
    _cuts = new Cut[0];
    _cutInverted = new boolean[0];
    _cutMap = new IdentityHashMap();
    _enabled = false;
    _inverted = false;
    _passThrough = true;
    _defaultEvent = new ChangeEvent(this);
  }

  public CutSet(CutSet set) {
    _cuts = new Cut[set._cuts.length];
    _cutInverted = new boolean[set._cuts.length];
    for (int i=0; i<_cuts.length; i++) {
      _cuts[i] = set._cuts[i];
      _cutInverted[i] = set._cutInverted[i];
    }
    _cutMap = new IdentityHashMap(set._cutMap);
    _enabled = set._enabled;
    _inverted = set._inverted;
    _passThrough = set._passThrough;
    _defaultEvent = new ChangeEvent(this);
  }

// -- Setters : ----------------------------------------------------------------

  /** Sets whether or not this CutSet is enabled. */
  public void setEnabled(boolean enabled) {
    boolean changed = (_enabled != enabled);
    _enabled = enabled;
    if (changed) fireStateChanged(_defaultEvent);
  }

  /** Sets whether or not this CutSet is inverted. */
  public void setInverted(boolean inverted) {
    boolean changed = (_inverted !=  inverted);
    _inverted =  inverted;
    if (changed) fireStateChanged(_defaultEvent);
  }

  /**
   * Sets whether or not this CutSet is in "pass through" mode.
   * When in "pass through" mode, inverting cuts allows objects that do not have attributes used by cuts to remain visible.
   * When not in "pass through" mode, inverting cuts forces objects that would be visible if the cuts were not inverted to become invisible.
   * If the cut is not inverted, "pass through" mode has no effect.
   */
  public void setPathThrough(boolean passThrough) {
    boolean changed = (_passThrough !=  passThrough);
    _passThrough =  passThrough;
    if (changed) fireStateChanged(_defaultEvent);
  }

  /** Sets whether or not the specified cut is enabled in this CutSet. */
  public void setEnabled(Cut cut, boolean enabled) {
    Boolean wasInverted = _cutMap.get(cut);
    if (wasInverted == null) {
      _cutMap.put(cut, false);
      cut.addCutSet(this);
      if (enabled) enable(cut);
    } else  {
      if (enabled) {
        enable(cut, wasInverted);
      } else {
        disable(cut);
      }
    }
  }

  /**
   * Sets whether or not the specified cut is inverted in this CutSet.
   * Implementation assumes that the specified cut is enabled.
   */
  public void setInverted(Cut cut, boolean inverted) {
    if (_cutMap.put(cut, inverted).booleanValue() != inverted) enable(cut, inverted);
  }

  /** 
   * Removes the specified cut from this CutSet.
   */
  public void remove(Cut cut) {
    if (_cutMap.remove(cut) != null) disable(cut);
  }

  protected void enable(Cut cut) {
    enable(cut, false);
  }

  protected void enable(Cut cut, boolean inverted) {
    _conditionNames = null;
    for (int i=0; i<_cuts.length; i++ ) {
      if (_cuts[i] == cut) {
        if (_cutInverted[i] != inverted) {
          _cutInverted[i] = inverted;
          fireStateChanged(_defaultEvent);
        }
        return;
      }
    }
    int len = _cuts.length;
    _cuts = Arrays.copyOf(_cuts, len+1);
    _cuts[len] = cut;
    _cutInverted = Arrays.copyOf(_cutInverted, len+1);
    _cutInverted[len] = inverted;
    fireStateChanged(_defaultEvent);
  }

  protected void disable(Cut cut) {
    _conditionNames = null;
    for (int i=0; i<_cuts.length; i++ ) {
      if (_cuts[i] == cut) {
        Cut[] cuts = new Cut[_cuts.length-1];
        boolean[] inv = new boolean[cuts.length];
        for (int iFrom=0, iTo=0; iFrom<_cuts.length; iFrom++) {
          if (i != iFrom) {
            cuts[iTo] = _cuts[iFrom];
            inv[iTo++] = _cutInverted[iFrom];
          }
        }
        _cuts = cuts;
        _cutInverted = inv;
        fireStateChanged(_defaultEvent);
        return;
      }
    }
  }
  
// -- Getters : ----------------------------------------------------------------

  /** Returns <tt>true</tt> if this CutSet is currently enabled. */
  public boolean isEnabled() {
    return _enabled;
  }

  /** Returns <tt>true</tt> if this CutSet is inverted. */
  public boolean isInverted() {
    return _inverted;
  }

  /** Returns <tt>true</tt> if this CutSet is in "pass through" mode. */
  public boolean isPassThrough() {
    return _passThrough;
  }

  /** Returns a set of all cuts this CutSet knows about - both enabled and disabled. */
  public Set<Cut> getCuts() {
    return Collections.unmodifiableSet(_cutMap.keySet());
  }

  /** Returns a set of names of all enabled conditions in this CutSet. */
  public Set<String> getConditionNames() {
    if (_conditionNames == null) {
      if (_cuts.length == 0) {
        _conditionNames = Collections.emptySet();
      } else {
        _conditionNames = new HashSet();
        for (int i = 0; i < size(); i++) {
          for (Condition c : _cuts[i].getConditions()) {
            if (c.isEnabled()) _conditionNames.add(c.getDefinition().getName());
          }
        }
      }
    }
    return _conditionNames;
  }

  /** Returns <tt>true</tt> if the specified cut is inverted in this CutSet. */
  public Boolean isCutInverted(Cut cut) {
    for (int i=0; i<_cuts.length; i++ ) {
      if (_cuts[i] == cut) return _cutInverted[i];
    }
    return null;
  }

  /** Returns <tt>true</tt> if the specified cut is enabled in this CutSet. */
  public Boolean isCutEnabled(Cut cut) {
    for (int i=0; i<_cuts.length; i++ ) {
      if (_cuts[i] == cut) return true;
    }
    return false;
  }

  /** Returns the number of enabled cuts in this CutSet. */
  public int size() {
    return _cuts.length;
  }


// -- Evaluation : -------------------------------------------------------------

  /**
   * Returns <tt>true</tt> if the supplied data satisfies this CutSet.
   * The returned value is determined by combining results returned by
   * {@link Cut#pass(Map,Object,boolean)} methods of all enabled cuts.
   * If this CutSet is not inverted, returns <tt>true</tt> if all cuts return either PASS or IGNORE.
   * If this CutSet is inverted and in "pass through" mode, returns <tt>true</tt> if either at least
   * one cut returns FAIL, or all cuts return IGNORE.
   * If this CutSet is inverted and not in "pass through" mode, returns <tt>true</tt> if at least one cut
   * returns FAIL.
   *
   * @param data  Map of attribute names to their values.
   * @param moreData  Additional data that can be used in making the decision -
   *                  ignored by the implementation provided by this class.
   */
  public boolean pass(Map<String,Value> data, Object moreData) {
    if (_inverted) {
      if (_passThrough) {  // inverted and pass through
        boolean out = true;
        for (int i = 0; i < _cuts.length; i++) {
          Cut.Result result = _cuts[i].pass(data, moreData, _cutInverted[i]);
          if (result == Cut.Result.FAIL) return true;
          else if (result == Cut.Result.PASS) out = false;
        }
        return out;
      } else {             // inverted and not path through
        for (int i = 0; i < _cuts.length; i++) {
          Cut.Result result = _cuts[i].pass(data, moreData, _cutInverted[i]);
          if (result == Cut.Result.FAIL) return true;
        }
        return false;
      }
    } else {               // not inverted
      for (int i = 0; i < _cuts.length; i++) {
        Cut.Result result = _cuts[i].pass(data, moreData, _cutInverted[i]);
        if (result == Cut.Result.FAIL) return false;
      }
      return true;
    }
  }

  /**
   * Returns <tt>true</tt> if the supplied data satisfies this CutSet.
   * Inversion and pass through modes are taken into account.
   *
   * @param data  Map of attribute names to their values.
   */
  public boolean pass(Map<String,Value> data) {
    return pass(data, null);
  }


// -- Event handling : ---------------------------------------------------------

  /** Adds a listener that will be notified of any changes in this CutSet, its enabled Cuts and Conditions. */
  public void addChangeListener(ChangeListener listener) {
    _listener = SwingEventMulticaster.add(_listener, listener);
  }

  /** Removes a listener. */
  public void removeChangeListener(ChangeListener listener) {
    _listener = SwingEventMulticaster.remove(_listener, listener);
  }

  /** Sends the event to listeners. */
  public void fireStateChanged(ChangeEvent event) {
    if (_listener != null) _listener.stateChanged(event);
  }

  /** Forwards the event to listeners if this CutSet is currently enabled. */
  public void stateChanged(ChangeEvent event) {
    _conditionNames = null;
    if (_enabled) fireStateChanged(event);
  }

// -----------------------------------------------------------------------------
}
