// Copyright 2005-2011, FreeHEP.
package hep.wired.cut;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.*;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JPopupMenu;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;

/**
 * GUI elements and logic for editing {@link CutSet}s.
 *
 * @author Dmitry Onoprienko
 * @version $Id: $
 */
public class CutSetEditor {

// -- Private parts : ----------------------------------------------------------

  private ArrayList<Cut> _cuts; // all known cuts
  private CutSet _cutSet; // CutSet currently being edited

  private CutEditor _cutEditor;

  private Action _actNew;
  private Action _actDelete;
  private Action _actLoad;
  private Action _actSave;

  private JCheckBox _enableBox;
  private JCheckBox _invertBox;
  private JCheckBox _passThroughBox;
  private JButton _actionsButton;
  private JTable _cutTable;
  private Model _cutTableModel;

// -- Construction and initialization : ----------------------------------------

  public CutSetEditor() {

    _cuts = new ArrayList();
    _cutEditor = new CutEditor();
    
    _actNew = new ActionNew();
    _actDelete = new ActionDelete();
    _actLoad = new AbstractAction("Load") {
      public void actionPerformed(ActionEvent e) {
        System.out.println("ActionLoad");
      }
    };
    _actSave = new AbstractAction("Save") {
      public void actionPerformed(ActionEvent e) {
        System.out.println("ActionSave");
      }
    };

    _enableBox = new JCheckBox("Apply");
    _enableBox.setToolTipText("Check to apply cuts to the current plot");
    _enableBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        boolean state = _enableBox.isSelected();
        if (_cutSet != null) _cutSet.setEnabled(state);
        setEnabled(state);
      }
    });

    _invertBox = new JCheckBox("Invert");
    _invertBox.setToolTipText("Check to invert cuts for the current plot");
    _invertBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        if (_cutSet != null) _cutSet.setInverted(_invertBox.isSelected());
        _passThroughBox.setEnabled(_invertBox.isSelected());
      }
    });

    _passThroughBox = new JCheckBox("Pass Through");
    _passThroughBox.setToolTipText("<html>If checked, objects that do not have attributes used by cuts remain visible whether or not the cut set is inverted<p>"
            + "If unchecked, inverting cuts hides all objects that would be visible if the cuts were not inverted");
    _passThroughBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        if (_cutSet != null) _cutSet.setPathThrough(_passThroughBox.isSelected());
      }
    });

    _cutTableModel = new Model();
    _cutTable = new JTable(_cutTableModel);
    _cutTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    _cutTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
      public void valueChanged(ListSelectionEvent e) {
        if (!e.getValueIsAdjusting()) {
          Cut cut = getSelectedCut();
          boolean hasSelection = (cut != null);
          _actDelete.setEnabled(hasSelection);
          _actSave.setEnabled(hasSelection);
          _cutEditor.setCut(cut);
          _cutEditor.setEnabled(hasSelection && _cutSet.isCutEnabled(cut));
        }
      }
    });

  }


// -- Setters : ----------------------------------------------------------------

  /**
   * Sets CutSet to be edited.
   * If the CutSet has cuts previously unknown to this editor, adds them.
   * If the CutSet does not have some of the cuts known ti this editor, adds them to the CutSet in disabled state.
   */
  public void setCutSet(CutSet cutSet) {
    _cutSet = cutSet;
    Set<Cut> cuts = cutSet.getCuts();
    for (Cut cut : cuts) {
      if (!_cuts.contains(cut)) _cuts.add(cut);
    }
    for (Cut cut : _cuts) {
      if (!cuts.contains(cut)) cutSet.setEnabled(cut, false);
    }
    _cutTableModel.fireTableDataChanged();
    _cutTable.clearSelection();
    _actDelete.setEnabled(false);
    _actSave.setEnabled(false);
    _cutEditor.setCut(null);
    _enableBox.setSelected(cutSet.isEnabled());
    _invertBox.setSelected(cutSet.isInverted());
    _passThroughBox.setSelected(cutSet.isPassThrough());
    setEnabled(cutSet.isEnabled());
  }

  public void setActionNew(Action action) {
    _actNew = action;
  }

  public void setActionDelete(Action action) {
    _actDelete = action;
  }

  public void setActionLoad(Action action) {
    _actLoad = action;
  }

  public void setActionSave(Action action) {
    _actSave = action;
  }

  public void setEnabled(boolean enabled) {
    _invertBox.setEnabled(enabled);
    _passThroughBox.setEnabled(enabled && _cutSet.isInverted());
    _actionsButton.setEnabled(enabled);
    if (!enabled) _cutTable.clearSelection();
  }


// -- Access to components : ---------------------------------------------------

  public CutEditor getCutEditor() {return _cutEditor;}

  public Action getActionNew() {return _actNew;}

  public Action getActionDelete() {return _actDelete;}

  public Action getActionLoad() {return _actLoad;}

  public Action getActionSave() {return _actSave;}

  public JCheckBox getEnableBox() {
    return _enableBox;
  }

  public JCheckBox getInvertBox() {
    return _invertBox;
  }

  public JCheckBox getPassThroughBox() {
    return _passThroughBox;
  }

  public JTable getCutTable() {
    return _cutTable;
  }

  public JButton getActionsButton() {
    final JPopupMenu pm = new JPopupMenu();
    pm.add(_actNew);
    pm.add(_actDelete);
//    pm.addSeparator();
//    pm.add(_actLoad);
//    pm.add(_actSave);
    _actionsButton = new JButton("Actions...");
    _actionsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        pm.show(_actionsButton, 0, _actionsButton.getHeight());
      }
    });
    return _actionsButton;
  }


// -- Getters : ----------------------------------------------------------------

  public Cut getSelectedCut() {
    int i = _cutTable.getSelectedRow();
    return i < 0 ? null : _cuts.get(_cutTable.convertRowIndexToModel(i));
  }

  public List<Cut> getCuts() {
    return Collections.unmodifiableList(_cuts);
  }

// -- Cuts table model class : -------------------------------------------------

  private class Model extends AbstractTableModel {

    public Class<?> getColumnClass(int columnIndex) {
      return columnIndex == 1 ? String.class : Boolean.class;
    }

    public String getColumnName(int column) {
      switch (column) {
        case 0:
          return "Apply";
        case 1:
          return "Name";
        default:
          return "Invert";
      }
    }

    public int getColumnCount() {
      return 3;
    }

    public int getRowCount() {
      return _cuts.size();
    }

    public Object getValueAt(int rowIndex, int columnIndex) {
      switch (columnIndex) {
        case 0:
          return _cutSet.isCutEnabled(_cuts.get(rowIndex));
        case 1:
          return _cuts.get(rowIndex).getName();
        case 2:
          return _cutSet.isCutInverted(_cuts.get(rowIndex));
      }
      throw new IllegalArgumentException();
    }

    public boolean isCellEditable(int rowIndex, int columnIndex) {
      return columnIndex == 2 ? _cutSet.isCutEnabled(_cuts.get(rowIndex)) : true;
    }

    public void setValueAt(Object value, int rowIndex, int columnIndex) {
      Cut cut = _cuts.get(rowIndex);
      switch (columnIndex) {
        case 0:
          boolean enabled = (Boolean)value;
          _cutSet.setEnabled(cut, enabled);
          _cutEditor.setEnabled(enabled);
          return;
        case 1:
          cut.setName((String)value);
          return;
        case 2:
          _cutSet.setInverted(cut, (Boolean)value);
      }
    }

  }

// -- Actions on CutSet : ------------------------------------------------------

  /**
   * Deletes the specified cut.
   * The cut is removed from the list maintained by this editor, and Cut's
   * {@link Cut#remove} method is called.
   */
  public void delete(Cut cut) {
    _cuts.remove(cut);
    _cutTableModel.fireTableDataChanged();
    _cutEditor.setCut(null);
    cut.remove();
  }

  /**
   * Adds the specified cut.
   * The cut is added to the list maintained by this editor and to the currently
   * edited {@link CutSet}. GUI is updated.
   */
  public void add(Cut cut) {
    _cuts.add(cut);
    _cutSet.setEnabled(cut, true);
    _cutTableModel.fireTableDataChanged();
    int row = _cutTable.convertRowIndexToView(_cuts.size() - 1);
    _cutTable.setRowSelectionInterval(row, row);
    _cutTable.editCellAt(row, _cutTable.convertColumnIndexToView(1));
    Component editor = _cutTable.getEditorComponent();
    try {
      JTextField tf = (JTextField) editor;
      tf.selectAll();
      tf.requestFocus();
    } catch (ClassCastException x) {}
  }

  public class ActionNew extends AbstractAction {
    public ActionNew() {
      super("New");
      putValue(SHORT_DESCRIPTION, "Create new cut");
    }
    public void actionPerformed(ActionEvent e) {
      add(new Cut());
    }
  }

  public class ActionDelete extends AbstractAction {
    public ActionDelete() {
      super("Delete");
      putValue(SHORT_DESCRIPTION, "Delete cut");
    }
    public void actionPerformed(ActionEvent e) {
      delete(getSelectedCut());
    }
  }

// -----------------------------------------------------------------------------
}
