// Copyright 2005-2011, FreeHEP.
package hep.wired.cut;

import org.freehep.util.Value;

/**
 * {@link Condition} based on comparison with <tt>double</tt> range.
 *
 * @author Mark Donszelmann
 * @author Dmitry Onoprienko
 * @version $Id: DoubleCondition.java 13824 2011-05-04 18:30:02Z onoprien $
 */
public class DoubleCondition extends AbstractCondition {

// -- Private parts : ----------------------------------------------------------

  double _min;
  double _max;

// -- Construction and initialization : ----------------------------------------

  public DoubleCondition(ConditionDefinition definition) {
    this(definition, false, false, 0., 0.);
  }

  public DoubleCondition(ConditionDefinition definition, boolean enabled, boolean inverted) {
    this(definition, enabled, inverted, 0., 0.);
  }

  public DoubleCondition(ConditionDefinition definition, double min, double max) {
    this(definition, false, false, min, max);
  }

  public DoubleCondition(ConditionDefinition definition, boolean enabled, boolean inverted, double min, double max) {
    super(definition, enabled, inverted);
    _min = min;
    _max = max;
  }


// -- Applying condition : -----------------------------------------------------

  /**
   * Returns <tt>true</tt> is the value satisfies this condition.
   * If the condition is not inverted, <tt>minimum &le; value &le; maximum</tt> must be satisfied.
   * If the condition is inverted, <tt>value &lt; minimum</tt> or <tt>maximum &lt; value</tt> must be satisfied.
   */
  public boolean pass(Value value) {
    if (_set) {
      double x = value.getDouble();
      return _inverted ? (x < _min || x > _max) :  (x >= _min && x <= _max);
    } else {
      return !_inverted;
    }
  }

// -- Setters : ----------------------------------------------------------------

  /**
   * Sets low bound of the acceptance range.
   * No consistency checks are applied.
   */
  public void setMinimum(double min) {
    _min = min;
    _set = true;
    fireStateChanged();
  }

  /**
   * Sets high bound of the acceptance range.
   * No consistency checks are applied.
   */
  public void setMaximum(double max) {
    _max = max;
    _set = true;
    fireStateChanged();
  }

  /**
   * Sets the acceptance range.
   * No consistency checks are applied.
   */
  public void setRange(double min, double max) {
    _min = min;
    _max = max;
    _set = true;
    fireStateChanged();
  }

  /**
   * Sets the acceptance range.
   * The actual type of the argument should be double[], otherwise <tt>ClassCastException</tt> is thrown.
   */
  public void set(Object value) {
    double[] range = (double[]) value;
    setRange(range[0], range[1]);
  }

// -- Getters : ----------------------------------------------------------------

  /** Returns status string describing what attribute values satisfy this condition in its current state. */
  public String getStatus() {
    if (!_set) return _inverted ? "none" : "all";
    if (_inverted) {
      return "x < "+ _min +" or "+ _max +" < x";
    } else {
      return _min +" <= x <= "+ _max;
    }
  }

  /** Returns low bound of the acceptance range. */
  public double getMinimum() {
    return _min;
  }

  /** Returns high bound of the acceptance range. */
  public double getMaximum() {
    return _max;
  }

  /** Returns the acceptance range. */
  public double[] get() {
    return new double[] {_min ,_max};
  }

// -----------------------------------------------------------------------------
}
