package hep.wired.cut;

/**
 *
 * @author Dmitry Onoprienko
 * @version $Id: $
 */
public final class IntegerSliderEditor extends SliderEditor {

// -- Private parts : ----------------------------------------------------------

  int _min, _max;  // Domain bounds as accessible by the slider
  private boolean _snap;

  private double _scale;
  private double _tolerance = .001;



// -- Construction and initialization : ----------------------------------------

  public IntegerSliderEditor(SliderPanel panel) {
    super(panel);
    _min = 1;
    _max = -1;
  }

// -- Setters : ----------------------------------------------------------------

  void setScale() {
    _scale = (_max - _min) / (double)(_slider._size);
  }

  void setDomain() {
    int[] domain = (int[]) getCondition().getDefinition().getDomain();
    int min = getCondition().getMinimum();
    int max = getCondition().getMaximum();
    _min = Math.min(domain[0], min);
    _max = Math.max(domain[1], max);
    if (_min == _max) {
      _min--;
      _max++;
    } else if (_min > _max) {
      int m = _min;
      _min = _max;
      _max = m;
    }
    if (_snap && min > max) _snap = false;
  }

// -- Updates from Slider : ----------------------------------------------------

  /** Called by ConditionSlider when its knobs have been moved. */
  void updateFromSlider(int low, int high, boolean valueIsAdjusting) {
    int min = 0;
    int max = 0;
    if (_snap) {
      min = (int) Math.round(_max - low * _scale);
      max = (int) Math.round(_max - high * _scale);
    } else {
      min = (int) Math.ceil(_max - low * _scale - _tolerance);
      max = (int) Math.floor(_max - high * _scale + _tolerance);
    }
    if (valueIsAdjusting == _applyImmediately) getCondition().setRange(min, max);
    if (valueIsAdjusting) {
      _slider.updateMarkPositions(convertToSlider(min), convertToSlider(max));
      _panel.getLowField().setText(convertToField(min));
      _panel.getHighField().setText(convertToField(max));
    } else {
      if (_snap) _slider.updateKnobPositions(convertToSlider(min), convertToSlider(max));
    }
  }

  /** Called by ConditionSlider when its low knob has been moved. */
  void updateFromSliderLow(int low, boolean valueIsAdjusting) {
    int min = (int) ( _snap ? Math.round(_max - low * _scale) : Math.ceil(_max - low * _scale - _tolerance) );
    if (valueIsAdjusting == _applyImmediately) getCondition().setMinimum(min);
    if (valueIsAdjusting) {
      _slider.updateMarkPositionLow(convertToSlider(min));
      _panel.getLowField().setText(convertToField(min));
    } else {
      if (_snap) _slider.updateKnobPositionLow(convertToSlider(min));
    }
  }

  /** Called by ConditionSlider when its high knob has been moved. */
  void updateFromSliderHigh(int high, boolean valueIsAdjusting) {
    int max = (int) ( _snap ? Math.round(_max - high * _scale) : Math.floor(_max - high * _scale + _tolerance) );
    if (valueIsAdjusting == _applyImmediately) getCondition().setMaximum(max);
    if (valueIsAdjusting) {
      _slider.updateMarkPositionHigh(convertToSlider(max));
      _panel.getHighField().setText(convertToField(max));
    } else {
      if (_snap) _slider.updateKnobPositionHigh(convertToSlider(max));
    }
  }


// -- Updates from panel : -----------------------------------------------------

  void setMark(boolean state) {
    _snap = state;
    if (_snap) {
      updateFromSlider(_slider._low, _slider._high, !_applyImmediately);
      updateFromSlider(_slider._low, _slider._high, _applyImmediately);
    }
  }

  void reset() {
    getCondition().reset();
    setDomain();
    _slider.doLayout();
    updateFields();
  }

  void updateFromLowField(String low) throws NumberFormatException {
    int min = Integer.parseInt(low);
    boolean _maxChanged = false;
    if (min > getCondition().getMaximum()) {
      getCondition().setRange(min, min);
      updateHighField();
      _maxChanged = true;
    } else {
      getCondition().setMinimum(min);
    }
    if (min < _min || min > _max) {
      setDomain();
      _slider.doLayout();
    } else {
      if (_maxChanged) {
        updateSlider();
      } else {
        updateSliderLow();
      }
    }
  }

  void updateFromHighField(String high) throws NumberFormatException {
    int max = Integer.parseInt(high);
    boolean _minChanged = false;
    if (max < getCondition().getMinimum()) {
      getCondition().setRange(max, max);
      updateLowField();
      _minChanged = true;
    } else {
      getCondition().setMaximum(max);
    }
    if (max > _max || max < _min) {
      setDomain();
      _slider.doLayout();
    } else {
      if (_minChanged) {
        updateSlider();
      } else {
        updateSliderLow();
      }
    }
  }


// -- Getters : ----------------------------------------------------------------

  public IntegerCondition getCondition() {
    return (IntegerCondition) _condition;
  }

  boolean getMark() throws IllegalArgumentException {
    return _snap;
  }

  /** Returns low bound of the range in slider coordinates. */
  int getLow() {
    return (int) Math.ceil((_max - getCondition().getMinimum()) / _scale);
  }

  /** Returns high bound of the range in slider coordinates. */
  int getHigh() {
    return (int) Math.floor((_max - getCondition().getMaximum()) / _scale);
  }

  /** Returns String representation of low bound. */
  String getLowString() {
    return convertToField(getCondition().getMinimum());
  }

  /** Returns String representation of high bound. */
  String getHighString() {
    return convertToField(getCondition().getMaximum());
  }

  
// -- Utility functions : ------------------------------------------------------

  int convertToSlider(int x) {
    return (int) Math.round((_max - x) / _scale);
  }

  String convertToField(int x) {
    return Integer.toString(x);
  }

// -- Creating ruller for Slider : ---------------------------------------------

  Slider.Ruler createRuler() {

    Slider.Ruler ruller = _slider.getNewRuler();

    int n = ruller.nDiv;

    int delta = (_max - _min) / n;
    int r = 1;
    while (r*10 < delta) r *= 10;
    if (r > 1) delta = delta - (delta % r);
    int digit = delta/r;
    int major = delta/2;
    int minor = delta/10;
    if (digit > 2 && digit < 8) {
      digit = 5;
      delta = digit * r;
      major = delta;
      minor = delta/5;
    } else if (digit > 7) {
      digit = 10;
      delta = digit * r;
      major = delta/2;
      minor = delta/10;
    } else if (digit == 2) {
      major = delta/2;
      minor = delta/10;
    }
    if (delta == 0) delta = 1;
    if (major == 0) major = delta;

    if (delta > 0) {
      int current = delta * (int) Math.ceil(_min / (double) delta);
      while (current <= _max) {
        int pos = convertToSlider(current);
        ruller.addLabel(pos, Integer.toString(current));
        current += delta;
      }
    } else {
      ruller.addLabel(convertToSlider(_min), Integer.toString(_min));
      if (_min != _max) ruller.addLabel(convertToSlider(_max), Integer.toString(_max));
    }

    if (major > 0) {
      int current = major * (int) Math.ceil(_min / (double) major);
      while (current <= _max) {
        int pos = convertToSlider(current);
        ruller.addMajorTick(pos);
        current += major;
      }
    } else {
      ruller.addMajorTick(convertToSlider(_min));
      if (_min != _max) ruller.addMajorTick(convertToSlider(_max));
    }

    if (minor > 0) {
      int current = minor * (int) Math.ceil(_min / (double) minor);
      while (current <= _max) {
        int pos = convertToSlider(current);
        ruller.addMinorTick(pos);
        current += minor;
      }
    }

    return ruller;
  }

// -----------------------------------------------------------------------------
}
