// Copyright 2005-2011, FreeHEP.
package hep.wired.cut;

import java.util.*;

import org.freehep.util.Value;

/**
 * {@link Condition} based on comparison with <tt>String</tt> set.
 *
 * @author Mark Donszelmann
 * @author Dmitry Onoprienko
 * @version $Id: StringCondition.java 13824 2011-05-04 18:30:02Z onoprien $
 */
public class StringCondition extends AbstractCondition {

// -- Private parts : ----------------------------------------------------------

  Set<String> _pass;

// -- Construction and initialization : ----------------------------------------

  public StringCondition(StringConditionDefinition definition) {
    this(definition, false, false, Collections.<String>emptySet());
  }

  public StringCondition(StringConditionDefinition definition, boolean enabled, boolean inverted) {
    this(definition, enabled, inverted, Collections.<String>emptySet());
  }

  public StringCondition(StringConditionDefinition definition, Collection<String> pass) {
    this(definition, false, false, pass);
  }

  public StringCondition(StringConditionDefinition definition, boolean enabled, boolean inverted, Collection<String> pass) {
    super(definition, enabled, inverted);
    _pass = pass.isEmpty() ? Collections.emptySet() : new HashSet(pass);
  }


// -- Applying condition : -----------------------------------------------------

  /**
   * Returns <tt>true</tt> is the value satisfies this condition.
   * If the condition is not inverted, <tt>value.getString()</tt> must be contained in 
   * String set associated with this condition.  If the condition is inverted, the value
   * must not be contained in the set.
   */
  public boolean pass(Value value) {
    return _inverted ^ _pass.contains(value.getString());
  }

// -- Setters : ----------------------------------------------------------------

  /** Sets the set of strings that satisfy this condition. */
  public void set(Collection<String> pass) {
    _pass = pass.isEmpty() ? Collections.emptySet() : new LinkedHashSet(pass);
    _set = true;
    fireStateChanged();
  }

  /**
   * Sets criteria to be used by this condition.
   * The actual type of the argument should be <tt>Collection&lt;String&gt;</tt>,
   * otherwise <tt>ClassCastException</tt> is thrown.
   */
  public void set(Object value) {
    set((Collection<String>)value);
  }

  /** Resets condition to its default state. */
  public void reset() {
    _pass = Collections.emptySet();
    _enabled = false;
    _inverted = false;
  }

// -- Getters : ----------------------------------------------------------------

  /** Returns <tt>ConditionDefinition</tt> associated with this condition. */
  public StringConditionDefinition getDefinition() {
    return (StringConditionDefinition) super.getDefinition();
  }

  /** Returns the set of strings that satisfy this condition. */
  public Set<String> get() {
    return Collections.unmodifiableSet(_pass);
  }

  /** Returns status string describing what attribute values satisfy this condition in its current state. */
  public String getStatus() {
    if (!_set || _pass.isEmpty()) return (_inverted ? "none" : "all");
    StringBuilder sb = new StringBuilder(_inverted ? "x not in [" : "x in [");
    for (String s : _pass) sb.append(s).append(", ");
    int size = sb.length();
    sb.delete(size-2, size).append("]");
    return sb.toString();
  }

// -----------------------------------------------------------------------------
}
