// Copyright 2004-2005, FreeHEP.
package hep.wired.edit;

import hep.wired.feature.Translateable;
import hep.wired.feature.Translateable2D;
import hep.wired.feature.Translateable3D;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.RecordPlot;
import hep.wired.util.UVWindices;

import java.awt.Component;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.util.logging.Logger;

import javax.swing.undo.UndoableEdit;

/**
 * Translates a plot in 2D or 3D.
 * 
 * @author Mark Donszelmann
 * @version $Id: Translate.java 8598 2006-08-14 20:38:02Z duns $
 */

public class Translate extends AnimatedWiredEdit implements GraphicsPanelEdit,
        UVWindices {

    private static Logger logger = Logger.getLogger(Translate.class
            .getPackage().getName());

    private double tx, ty, tz;

    /**
     * Creates a translate edit with no translation.
     */
    public Translate() {
        this(0, 0, 0);
    }

    /**
     * Creates a translate edit with given translation in model units.
     */
    public Translate(double tx, double ty, double tz) {
        this(tx, ty, tz, null, 0);
    }

    /**
     * Creates a translate edit with given translation in model units, animation
     * shape and number of frames.
     */
    public Translate(double tx, double ty, double tz, Shape shape, int frames) {
        super(shape, frames);
        this.tx = tx;
        this.ty = ty;
        this.tz = tz;
    }

    public WiredEdit copy(RecordPlot plot) {
        WiredEdit copy = new Translate(tx, ty, tz, getShape(), getFrames());
        copy.setRecordPlot(plot);
        return copy;
    }

    protected Shape[] getShapes(Shape shape, int steps) {
        if ((shape == null) || (steps <= 1))
            return null;

        Shape[] shapes = new Shape[steps];

        // calculate translate increments
        GraphicsPanel panel = getRecordPlot().getGraphicsPanel();
        Translateable translateable = (Translateable) panel
                .getFeature(Translateable.class);
        double[] uvw = translateable.getScreenTranslation(new double[] {
                tx / (steps - 1), ty / (steps - 1), tz / (steps - 1) }, panel
                .getViewPort());
        AffineTransform transform = new AffineTransform();

        // calculate steps
        for (int i = 0; i < steps; i++) {
            // transform
            shapes[i] = transform.createTransformedShape(shape);
            // add our increment
            transform.translate(uvw[U], -uvw[V]);
        }
        return shapes;
    }

    public Shape createTransformedShape(Component component, Shape shape) {
        RecordPlot plot = (RecordPlot) component;
        GraphicsPanel panel = plot.getGraphicsPanel();
        Translateable translateable = (Translateable) panel
                .getFeature(Translateable.class);
        double[] uvw = translateable.getScreenTranslation(new double[] { tx,
                ty, tz }, panel.getViewPort());
        return AffineTransform.getTranslateInstance(uvw[U], -uvw[V])
                .createTransformedShape(shape);
    }

    public String getPresentationName() {
        StringBuffer s = new StringBuffer();
        s.append(toString());
        s.append(" (");
        s.append(format.format(tx));
        s.append(", ");
        s.append(format.format(ty));
        if (tz != 0) {
            s.append(", ");
            s.append(format.format(tz));
        }
        s.append(")");
        return s.toString();
    }

    public String toString() {
        return "Translate";
    }

    public boolean addEdit(UndoableEdit edit) {
        if (edit instanceof Translate) {
            Translate translate = (Translate) edit;
            tx += translate.tx;
            ty += translate.ty;
            tz += translate.tz;
            return true;
        }
        return false;
    }

    public boolean isSupportedBy(GraphicsPanel p) {
        return p.supports(Translateable2D.class)
                || p.supports(Translateable3D.class);
    }

    protected void redoEdit() {
        GraphicsPanel panel = getRecordPlot().getGraphicsPanel();
        Translateable3D translateable3D = (Translateable3D) panel
                .getFeature(Translateable3D.class);
        if (translateable3D != null) {
            translateable3D.translate(tx, ty, tz);
        } else {
            Translateable2D translateable2D = (Translateable2D) panel
                    .getFeature(Translateable2D.class);
            if (translateable2D == null) {
                System.err.println("Could not redo: " + this);
                return;
            }
            translateable2D.translate(tx, ty);
        }
        getRecordPlot().repaint();
        logger.finer(toString());
    }

    protected void undoEdit() {
        GraphicsPanel panel = getRecordPlot().getGraphicsPanel();
        Translateable3D translateable3D = (Translateable3D) panel
                .getFeature(Translateable3D.class);
        if (translateable3D != null) {
            translateable3D.translate(-tx, -ty, -tz);
        } else {
            Translateable2D translateable2D = (Translateable2D) panel
                    .getFeature(Translateable2D.class);
            if (translateable2D == null) {
                System.err.println("Could not undo: " + this);
                return;
            }
            translateable2D.translate(-tx, -ty);
        }
        getRecordPlot().repaint();
        logger.finer(toString());
    }
}
