package hep.wired.interaction;

import hep.wired.edit.Scale;
import hep.wired.edit.Translate;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.RecordPlot;

import java.awt.Cursor;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;

/**
 * Default implementation of {@link InteractionHandler} that handles mouse wheel
 * events to scroll, cursor keys to translate.
 *
 * @author Mark Donszelmann
 * @version $Id: DefaultInteractionHandler.java 14093 2013-01-09 21:57:40Z onoprien $
 */
public class DefaultInteractionHandler extends AbstractInteractionHandler {

  private static DefaultInteractionHandler instance;
  
  protected static final double fixedScale = Math.sqrt(2);
  protected static final double multiplyScale = 3;
  protected static final double fixedTranslate = 1000;
  protected static final double multiplyTranslate = 10;
  
  private double lastScale;
  
  private static boolean mac = System.getProperty("os.name").toLowerCase().indexOf("mac") != -1;
  public static final String command = "\u2318";      // four-clover unicode on MacOS X

  /**
   * Creates a DefaultInteractionHandler.
   */
  protected DefaultInteractionHandler(String name) {
    super(name);
    lastScale = 1.0;
  }

  public static DefaultInteractionHandler getInstance() {
    if (instance == null) instance = new DefaultInteractionHandler("Default");
    return instance;
  }

  public String getDescription() {
    return "Move mousewheel to zoom in/out; Use cursor keys to translate.";
  }

  public boolean isSupportedBy(GraphicsPanel panel) {
    return true;
  }

  public void changeCursor(RecordPlot plot, InputEvent event) {
    plot.setCursor(Cursor.getDefaultCursor());
  }

  public void mouseEntered(RecordPlot plot, MouseEvent event) {
    plot.requestFocusInWindow();
    changeCursor(plot, event);
  }

  protected double getLastScale() {
    return lastScale;
  }

  public void mouseWheelMoved(RecordPlot plot, MouseWheelEvent event) {
    if (!plot.supports(new Scale())) {
      lastScale = 1;
      return;
    }

    double scale = fixedScale - 1.0;
    scale *= event.isShiftDown() ? multiplyScale : 1;
    scale *= event.isControlDown() ? 1 / multiplyScale : 1;
    scale += 1.0;
    double clicks = event.getWheelRotation();
    scale *= clicks;
    scale = (scale < 0) ? 1 / -scale : scale;
    lastScale = scale;
    plot.postEdit(new Scale(scale, scale, scale));
  }

  public boolean otherKeyPressed(RecordPlot plot, KeyEvent event) {
    switch (event.getKeyCode()) {
      case KeyEvent.VK_UP:
      case KeyEvent.VK_KP_UP:
        translate(plot, event, 0, fixedTranslate);
        return true;

      case KeyEvent.VK_DOWN:
      case KeyEvent.VK_KP_DOWN:
        translate(plot, event, 0, -fixedTranslate);
        return true;

      case KeyEvent.VK_LEFT:
      case KeyEvent.VK_KP_LEFT:
        translate(plot, event, -fixedTranslate, 0);
        return true;

      case KeyEvent.VK_RIGHT:
      case KeyEvent.VK_KP_RIGHT:
        translate(plot, event, fixedTranslate, 0);
        return true;

      case KeyEvent.VK_PLUS:
      case KeyEvent.VK_ADD:
        if (plot.supports(new Scale())) {
          plot.postEdit(new Scale(fixedScale, fixedScale, fixedScale));
        }
        return true;

      case KeyEvent.VK_MINUS:
      case KeyEvent.VK_SUBTRACT:
        if (plot.supports(new Scale())) {
          plot.postEdit(new Scale(1 / fixedScale, 1 / fixedScale,
                  1 / fixedScale));
        }
        return true;

      default:
        return false;
    }
  }

  public static boolean isControlMetaDown(InputEvent event) {
    return isMac() ? event.isMetaDown() : event.isControlDown();
  }

  public static boolean isMac() {
    return mac;
  }

  private void translate(RecordPlot plot, InputEvent event, double dx,
          double dy) {
    if (!plot.supports(new Translate())) {
      return;
    }

    if (event.isControlDown()) {
      dx /= multiplyTranslate;
      dy /= multiplyTranslate;
    }

    if (event.isShiftDown()) {
      dx *= multiplyTranslate;
      dy *= multiplyTranslate;
    }

    plot.postEdit(new Translate(dx, dy, 0));
  }
}
