// Copyright 2004-2006, FreeHEP.
package hep.wired.interaction;

import hep.wired.edit.Scale;
import hep.wired.image.WiredBaseImage;
import hep.wired.services.GraphicsPanel;
import hep.wired.services.RecordPlot;

import java.awt.Shape;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import javax.swing.Icon;

import org.freehep.application.Application;

/**
 * Drag/click mouse to scale plot.
 * 
 * @author Mark Donszelmann
 * @version $Id: DragToScale.java 8598 2006-08-14 20:38:02Z duns $
 */
public class DragToScale extends DefaultInteractionHandler {

    private double dpsq;

    /**
     * Create a drag handler to generate scale edits.
     */
    public DragToScale() {
        super("Drag to Scale");
    }

    public boolean isSupportedBy(GraphicsPanel panel) {
        return new Scale().isSupportedBy(panel);
    }

    public Icon getIcon(int size) {
        return WiredBaseImage.getIcon("Scale%w", size);
    }

    public String getDescription() {
        return "Click to zoom in, Shift-Click to zoom out, Drag inward or outward to instant zoom.";
    }

    public void changeCursor(RecordPlot plot, InputEvent event) {
        if (event.getID() == MouseEvent.MOUSE_DRAGGED) {
            plot.setCursor(WiredBaseImage
                    .getBestCursor("ScaleCursor%w", 32, 32));
        } else if (event.isShiftDown()) {
            plot.setCursor(WiredBaseImage.getBestCursor("ScaleMinusCursor%w",
                    32, 32));
        } else {
            plot.setCursor(WiredBaseImage.getBestCursor("ScalePlusCursor%w",
                    32, 32));
        }
    }

    public void reset(RecordPlot plot, InputEvent event) {
        plot.getGraphicsPanel().setFastMode(false);
        plot.drawShape(null);
        plot.repaint();
        Application.getApplication().setStatusMessage(getDescription());
    }

    public void mouseEntered(RecordPlot plot, MouseEvent event) {
        plot.requestFocusInWindow();
        changeCursor(plot, event);
    }

    public void mouseButton1DragStarted(RecordPlot plot, MouseEvent event) {
        dpsq = Point2D.distanceSq(event.getComponent().getWidth() / 2.0, event
                .getComponent().getHeight() / 2.0, event.getX(), event.getY());
        changeCursor(plot, event);
        plot.beginUpdate();
        plot.getGraphicsPanel().setFastMode(true);
    }

    public void mouseButton1Dragged(RecordPlot plot, MouseEvent event) {
        double ddsq = Point2D.distanceSq(event.getComponent().getWidth() / 2.0,
                event.getComponent().getHeight() / 2.0, event.getX(), event
                        .getY());
        double s = Math.sqrt(ddsq / dpsq);

        plot.postEdit(new Scale(s, s, s));
        dpsq = ddsq;
        defineShape(plot, event);
    }

    public void mouseButton1DragEnded(RecordPlot plot, MouseEvent event) {
        plot.endUpdate();
        reset(plot, event);
    }

    public void mouseButton1Released(RecordPlot plot, MouseEvent event) {
        changeCursor(plot, event);
    }

    public boolean altKeyPressed(RecordPlot plot, KeyEvent event) {
        changeCursor(plot, event);
        return true;
    }

    public boolean altKeyReleased(RecordPlot plot, KeyEvent event) {
        changeCursor(plot, event);
        return true;
    }

    /**
     * Zoom in (or out if alt key pressed).
     */
    public void mouseButton1Clicked(RecordPlot plot, MouseEvent event) {
        double w = event.getComponent().getWidth();
        double h = event.getComponent().getHeight();

        plot.beginUpdate();
        // zoom
        if (event.isShiftDown()) {
            // zoom out
            Shape shape = new Rectangle2D.Double(0, 0, w, h);
            plot.postEdit(new Scale(1 / fixedScale, 1 / fixedScale,
                    1 / fixedScale, shape, 5));
        } else {
            // zoom in
            Shape shape = new Rectangle2D.Double((w - (w / fixedScale)) / 2,
                    (h - (h / fixedScale)) / 2, w / fixedScale, h / fixedScale);
            plot.postEdit(new Scale(fixedScale, fixedScale, fixedScale, shape,
                    5));
        }
        plot.endUpdate();
        reset(plot, event);
    }

    private void defineShape(RecordPlot plot, MouseEvent event) {
        double x = event.getComponent().getWidth() / 2.0;
        double y = event.getComponent().getHeight() / 2.0;
        double d = Point2D.distance(event.getX(), event.getY(), x, y);
        plot.drawShape(new Ellipse2D.Double(x - d, y - d, d * 2, d * 2));
    }

    public String toString() {
        return "Drag to Scale";
    }
}
