// Copyright 2004-2006, FreeHEP.
package hep.wired.plot;

import hep.wired.edit.FitToWindow;
import hep.wired.edit.Reset;
import hep.wired.interaction.DragAngledRectangleToScale;
import hep.wired.interaction.DragBallToRotate;
import hep.wired.interaction.DragRectangleToScale;
import hep.wired.interaction.DragToRotate;
import hep.wired.interaction.DragToScale;
import hep.wired.interaction.DragToTranslate;
import hep.wired.services.RecordPlot;
import hep.wired.util.WiredUndoManager;

import java.awt.Component;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;

import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;

import org.freehep.application.Application;
import org.freehep.application.PrintHelper;
import org.freehep.application.studio.Studio;
import org.freehep.util.commanddispatcher.BooleanCommandState;
import org.freehep.util.commanddispatcher.CommandProcessor;
import org.freehep.util.commanddispatcher.CommandState;
import org.freehep.graphicsbase.util.export.VectorGraphicsTransferable;

/**
 * Defines a command handler for plot commands.
 * 
 * @author Mark Donszelmann
 * @version $Id: WiredPlotCommandHandler.java 14047 2012-10-22 20:35:37Z onoprien $
 */
public class WiredPlotCommandHandler extends CommandProcessor implements
        UndoableEditListener {

    private RecordPlot plot;

    private WiredUndoManager undoManager = new WiredUndoManager();

    /**
     * Creates a command handler attached to the plot.
     */
    public WiredPlotCommandHandler(RecordPlot plot) {
        super();
        this.plot = plot;
    }

    public void enableRedo(CommandState state) {
        state.setEnabled(undoManager.canRedo());
        state.setText("Redo " + undoManager.getRedoString());
        state.setToolTipText(undoManager.getRedoPresentationName());
        setChanged();
    }

    public void onRedo() {
        undoManager.redo();
        setChanged();
    }

    public void enableUndo(CommandState state) {
        state.setEnabled(undoManager.canUndo());
        state.setText("Undo " + undoManager.getUndoString());
        state.setToolTipText(undoManager.getUndoPresentationName());
        setChanged();
    }

    public void onUndo() {
        undoManager.undo();
        setChanged();
    }

    public void onPrintSetup() {
        Studio studio = (Studio) Application.getApplication();

        PrintHelper printHelper = new PrintHelper((Component) plot, studio);
        printHelper.showOptionsDialog(studio);
    }

    public void enablePrint(CommandState state) {
        state.setEnabled(plot.getRecord() != null);
        setChanged();
    }

    public void onPrint() throws Exception {
        Studio studio = (Studio) Application.getApplication();

        PrintHelper printHelper = new PrintHelper((Component) plot, studio);
        printHelper.print();
    }

    public void enablePrintPreview(CommandState state) {
        state.setEnabled(plot.getRecord() != null);
        setChanged();
    }

    public void onPrintPreview() throws Exception {
        Studio studio = (Studio) Application.getApplication();

        PrintHelper printHelper = new PrintHelper((Component) plot, studio);
        printHelper.printPreview(studio);
    }

    public void enableCopy(CommandState state) {
        state.setEnabled(plot.getRecord() != null);
        setChanged();
    }

    public void onCopy() {
        Clipboard clipBoard = Toolkit.getDefaultToolkit().getSystemClipboard();
        VectorGraphicsTransferable transferable = new VectorGraphicsTransferable(
                (Component) plot);
        clipBoard.setContents(transferable, transferable);
    }

    public void undoableEditHappened(UndoableEditEvent edit) {
        undoManager.addEdit(edit.getEdit());
        setChanged();
    }

    public void enableDragToScale(BooleanCommandState state) {
        state.setEnabled(plot.supports(new DragToScale()));
        state.setSelected(plot.getInteractionHandler() instanceof DragToScale);
        setChanged();
    }

    public void onDragToScale(boolean on) {
        if (on)
            plot.setInteractionHandler(new DragToScale());
        setChanged();
    }

    public void enableDragRectangleToScale(BooleanCommandState state) {
        state.setEnabled(plot.supports(new DragRectangleToScale()));
        state
                .setSelected(plot.getInteractionHandler() instanceof DragRectangleToScale);
        setChanged();
    }

    public void onDragRectangleToScale(boolean on) {
        if (on)
            plot.setInteractionHandler(new DragRectangleToScale());
        setChanged();
    }

    public void enableDragAngledRectangleToScale(BooleanCommandState state) {
        state.setEnabled(plot.supports(new DragAngledRectangleToScale()));
        state
                .setSelected(plot.getInteractionHandler() instanceof DragAngledRectangleToScale);
        setChanged();
    }

    public void onDragAngledRectangleToScale(boolean on) {
        if (on)
            plot.setInteractionHandler(new DragAngledRectangleToScale());
        setChanged();
    }

    public void enableDragToTranslate(BooleanCommandState state) {
        state.setEnabled(plot.supports(new DragToTranslate()));
        state
                .setSelected(plot.getInteractionHandler() instanceof DragToTranslate);
        setChanged();
    }

    public void onDragToTranslate(boolean on) {
        if (on)
            plot.setInteractionHandler(new DragToTranslate());
        setChanged();
    }

    public void enableDragToRotate(BooleanCommandState state) {
        state.setEnabled(plot.supports(new DragToRotate()));
        state.setSelected(plot.getInteractionHandler() instanceof DragToRotate);
        setChanged();
    }

    public void onDragToRotate(boolean on) {
        if (on)
            plot.setInteractionHandler(new DragToRotate());
        setChanged();
    }

    public void enableDragBallToRotate(BooleanCommandState state) {
        state.setEnabled(plot.supports(new DragBallToRotate()));
        state
                .setSelected(plot.getInteractionHandler() instanceof DragBallToRotate);
        setChanged();
    }

    public void onDragBallToRotate(boolean on) {
        if (on)
            plot.setInteractionHandler(new DragBallToRotate());
        setChanged();
    }

    public void enableReset(CommandState state) {
        state.setEnabled(plot.supports(new Reset()));
        setChanged();
    }

    public void onReset() {
        plot.postEdit(new Reset());
    }

    public void enableFitToWindow(CommandState state) {
        state.setEnabled(plot.supports(new FitToWindow()));
        setChanged();
    }

    public void onFitToWindow() {
        plot.postEdit(new FitToWindow());
    }

}
