// Copyright 2004-2005, FreeHEP.
package hep.wired.util;

import java.awt.geom.NoninvertibleTransformException;
import java.text.DecimalFormat;

import org.freehep.xml.io.XMLIO;
import org.freehep.xml.io.XMLIOManager;
import org.jdom.DataConversionException;

/**
 * A 2x3 matrix for projection of 2D space to 2D space.
 * 
 * @author Mark Donszelmann
 * @version $Id: Matrix2D.java 8598 2006-08-14 20:38:02Z duns $
 */

public class Matrix2D implements XYZindices, UVWindices, Cloneable, XMLIO {

    private static final DecimalFormat fmt = new DecimalFormat("0000.000");

    private double m00, m01, m02;

    private double m10, m11, m12;

    /**
     * Creates an identity matrix.
     */
    public Matrix2D() {
        this(1, 0, 0, 1, 0, 0);
    }

    /**
     * Creates a matrix with given parameters.
     * @param m00 matrix param 00
     * @param m10 matrix param 10
     * @param m01 matrix param 01
     * @param m11 matrix param 11
     * @param m02 matrix param 02
     * @param m12 matrix param 12
     */
    public Matrix2D(double m00, double m10, double m01, double m11, double m02,
            double m12) {
        set(m00, m10, m01, m11, m02, m12);
    }

    public Object clone() {
        return new Matrix2D(m00, m10, m01, m11, m02, m12);
    }

    public int hashCode() {
        long bits;
        bits = Double.doubleToLongBits(m00);
        bits = bits * 31 + Double.doubleToLongBits(m10);
        bits = bits * 31 + Double.doubleToLongBits(m01);
        bits = bits * 31 + Double.doubleToLongBits(m11);
        bits = bits * 31 + Double.doubleToLongBits(m02);
        bits = bits * 31 + Double.doubleToLongBits(m12);
        return (((int) bits) ^ ((int) (bits >> 32)));
    }

    public boolean equals(Object obj) {
        if (!(obj instanceof Matrix2D))
            return super.equals(obj);

        Matrix2D m = (Matrix2D) obj;

        return (m.m00 == m00) && (m.m10 == m10) && (m.m01 == m01)
                && (m.m11 == m11) && (m.m02 == m02) && (m.m12 == m12);
    }

    /**
     * @return the sx value (m00).
     */
    public double getScaleX() {
        return m00;
    }

    /**
     * @return the sy value (m11).
     */
    public double getScaleY() {
        return m11;
    }

    /**
     * @return the shx value (m01).
     */
    public double getShearX() {
        return m01;
    }

    /**
     * @return the shy value (m10).
     */
    public double getShearY() {
        return m10;
    }

    /**
     * @return the tx value (m02).
     */
    public double getTranslateX() {
        return m02;
    }

    /**
     * @return the ty value (m12).
     */
    public double getTranslateY() {
        return m12;
    }

    private void set(double m00, double m10, double m01, double m11,
            double m02, double m12) {
        this.m00 = m00;
        this.m10 = m10;
        this.m01 = m01;
        this.m11 = m11;
        this.m02 = m02;
        this.m12 = m12;
    }

    /**
     * @return the determinant.
     */
    public double getDeterminant() {
        return m00 * m11 - m01 * m10;
    }

    /**
     * @return the inverse of this matrix.
     * @throws NoninvertibleTransformException if matrix cannot be inverted
     */
    public Matrix2D createInverse() throws NoninvertibleTransformException {
        double d = getDeterminant();
        if (Math.abs(d) <= Double.MIN_VALUE)
            throw new NoninvertibleTransformException(
                    "Matrix2D cannot be inverted. Determinant: " + d + "\n"
                            + toString());

        return new Matrix2D(m11 / d, -m10 / d, -m01 / d, m00 / d,
                (m01 * m12 - m11 * m02) / d, (m10 * m02 - m00 * m12) / d);
    }

    /**
     * Concatenates this matrix with m, such that M' = M * m.
     * @param m post matrix
     */
    public void concatenate(Matrix2D m) {
        set(m00 * m.m00 + m01 * m.m10, m10 * m.m00 + m11 * m.m10,

        m00 * m.m01 + m01 * m.m11, m10 * m.m01 + m11 * m.m11,

        m00 * m.m02 + m01 * m.m12 + m02, m10 * m.m02 + m11 * m.m12 + m12);
    }

    /**
     * Pre-concatenates this matrix with m, such that M' = m * M.
     * @param m pre matrix
     */
    public void preConcatenate(Matrix2D m) {
        set(m.m00 * m00 + m.m01 * m10, m.m10 * m00 + m.m11 * m10,

        m.m00 * m01 + m.m01 * m11, m.m10 * m01 + m.m11 * m11,

        m.m00 * m02 + m.m01 * m12 + m.m02, m.m10 * m02 + m.m11 * m12 + m.m12);
    }

    /**
     * Rotates theta radians. Preconcatenates with the matrix:
     * 
     * <pre>
     *   [   cos(theta)      -sin(theta)     0       ]
     *   [   sin(theta)      cos(theta)      0       ]
     *   [   0               0               1       ]
     * </pre>
     * @param theta angle
     */
    public void rotate(double theta) {
        // cos*m00-sin*m10 cos*m01-sin*m11 cos*m02-sin*m12
        // sin*m00+cos*m10 sin*m01+cos*m11 sin*m02+cos+m12
        double cos = Math.cos(theta);
        double sin = Math.sin(theta);
        set(cos * m00 - sin * m10, sin * m00 + cos * m10,
                cos * m01 - sin * m11, sin * m01 + cos * m11, cos * m02 - sin
                        * m12, sin * m02 + cos * m12);
    }

    /**
     * Scales by sx and sy. Preconcatenates with the matrix:
     * 
     * <pre>
     *   [   sx              0               0       ]
     *   [   0               sy              0       ]
     *   [   0               0               1       ]
     * </pre>
     * @param sx scale factor in x
     * @param sy scale factor in y
     */
    public void scale(double sx, double sy) {
        // sx*m00 sx*m01 sx*m02
        // sy*m10 sy*m11 sy*m12
        m00 *= sx;
        m01 *= sx;
        m02 *= sx;
        m10 *= sy;
        m11 *= sy;
        m12 *= sy;
    }

    /**
     * Shears by shx and shy. Preconcatenates with the matrix:
     * 
     * <pre>
     *   [   1               shx             0       ]
     *   [   shy             1               0       ]
     *   [   0               0               1       ]
     * </pre>
     * @param shx shear factor in x
     * @param shy shear factor in y
     */
    public void shear(double shx, double shy) {
        // m00+shx*m10 m01+shx*m11 m02+shx*m12
        // m10+shy*m00 m11+shy*m01 m12+shy*m02
        set(m00 + shx * m10, m10 + shy * m00, m01 + shx * m11, m11 + shy * m01,
                m02 + shx * m12, m12 + shy * m02);
    }

    /**
     * Translates by tx and ty. Preconcatenates with the matrix:
     * 
     * <pre>
     *   [   1               0               tx      ]
     *   [   0               1               ty      ]
     *   [   0               0               1       ]
     * </pre>
     * @param tx translation in x
     * @param ty translation in y
     */
    public void translate(double tx, double ty) {
        // m00 m01 m02+tx
        // m10 m11 m12+ty
        m02 += tx;
        m12 += ty;
    }

    /**
     * Model-Translates by tx and ty. Concatenates with the matrix:
     * 
     * <pre>
     *   [   1               0               tx      ]
     *   [   0               1               ty      ]
     *   [   0               0               1       ]
     * </pre>
     * @param tx translation in x
     * @param ty translation in y
     */
    public void modelTranslate(double tx, double ty) {
        // m00 m01 m00*tx+m01*ty+m02
        // m10 m11 m10*tx+m11*ty+m12
        m02 += m00 * tx + m01 * ty;
        m12 += m10 * tx + m11 * ty;
    }

    /**
     * Transforms xyz by this matrix, using translation if delta is false. The
     * input array is returned.
     * @param xyz input xyz
     * @param delta use translation
     * @return uvw
     */
    public double[] transform(double[] xyz, boolean delta) {
        int d = delta ? 0 : 1;
        double x = xyz[X];
        double y = xyz[Y];

        xyz[U] = m00 * x + m01 * y + m02 * d;
        xyz[V] = m10 * x + m11 * y + m12 * d;

        return xyz;
    }

    /**
     * Transforms xyz by this matrix for n points, using translation if delta is
     * false. The input array is returned.
     * @param xyz input xyz
     * @param n number of points
     * @param delta use translation
     * @return uvw
     */
    public double[][] transform(double[][] xyz, int n, boolean delta) {
        int d = delta ? 0 : 1;
        for (int i = 0; i < n; i++) {
            double x = xyz[X][i];
            double y = xyz[Y][i];
            xyz[U][i] = m00 * x + m01 * y + m02 * d;
            xyz[V][i] = m10 * x + m11 * y + m12 * d;
        }
        return xyz;
    }

    public String toString() {
        StringBuffer s = new StringBuffer();
        s.append("Matrix2D: ");
        s.append(fmt.format(m00));
        s.append(", ");
        s.append(fmt.format(m01));
        s.append(", ");
        s.append(fmt.format(m02));
        s.append("\n");

        s.append("          ");
        s.append(fmt.format(m10));
        s.append(", ");
        s.append(fmt.format(m11));
        s.append(", ");
        s.append(fmt.format(m12));
        s.append("\n");

        return s.toString();
    }

    //
    // XMLIO
    //
    public void save(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        nodeEl.setAttribute("m00", Double.toString(m00));
        nodeEl.setAttribute("m10", Double.toString(m10));

        nodeEl.setAttribute("m01", Double.toString(m01));
        nodeEl.setAttribute("m11", Double.toString(m11));

        nodeEl.setAttribute("m02", Double.toString(m02));
        nodeEl.setAttribute("m12", Double.toString(m12));
    }

    public void restore(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        try {
            set(nodeEl.getAttribute("m00").getDoubleValue(), nodeEl
                    .getAttribute("m10").getDoubleValue(),

            nodeEl.getAttribute("m01").getDoubleValue(), nodeEl.getAttribute(
                    "m11").getDoubleValue(),

            nodeEl.getAttribute("m02").getDoubleValue(), nodeEl.getAttribute(
                    "m12").getDoubleValue());
        } catch (DataConversionException dce) {
            throw new IllegalArgumentException(getClass() + ": "
                    + dce.getMessage());
        }
    }
}
