// Copyright 2004-2005, FreeHEP.
package hep.wired.util;

import java.awt.geom.NoninvertibleTransformException;
import java.text.DecimalFormat;

import org.freehep.xml.io.XMLIO;
import org.freehep.xml.io.XMLIOManager;
import org.jdom.DataConversionException;

/**
 * A 3x4 matrix for projections of 3D space to 3D space.
 * 
 * @author Mark Donszelmann
 * @version $Id: Matrix3D.java 8598 2006-08-14 20:38:02Z duns $
 */

public class Matrix3D implements XYZindices, UVWindices, Cloneable, XMLIO {

    private static final DecimalFormat fmt = new DecimalFormat("0000.000");

    private double m00, m01, m02, m03;

    private double m10, m11, m12, m13;

    private double m20, m21, m22, m23;

    private boolean orthoNormalized;

    /**
     * Creates an identity matrix.
     */
    public Matrix3D() {
        this(1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0);
    }

    /**
     * Creates a matrix with given parameters.
     */
    public Matrix3D(double m00, double m10, double m20, double m01, double m11,
            double m21, double m02, double m12, double m22, double m03,
            double m13, double m23) {
        set(m00, m10, m20, m01, m11, m21, m02, m12, m22, m03, m13, m23);
    }

    public Object clone() {
        return new Matrix3D(m00, m10, m20, m01, m11, m21, m02, m12, m22, m03,
                m13, m23);
    }

    public int hashCode() {
        long bits;
        bits = Double.doubleToLongBits(m00);
        bits = bits * 31 + Double.doubleToLongBits(m10);
        bits = bits * 31 + Double.doubleToLongBits(m20);
        bits = bits * 31 + Double.doubleToLongBits(m01);
        bits = bits * 31 + Double.doubleToLongBits(m11);
        bits = bits * 31 + Double.doubleToLongBits(m21);
        bits = bits * 31 + Double.doubleToLongBits(m02);
        bits = bits * 31 + Double.doubleToLongBits(m12);
        bits = bits * 31 + Double.doubleToLongBits(m22);
        bits = bits * 31 + Double.doubleToLongBits(m03);
        bits = bits * 31 + Double.doubleToLongBits(m13);
        bits = bits * 31 + Double.doubleToLongBits(m23);
        return (((int) bits) ^ ((int) (bits >> 32)));
    }

    public boolean equals(Object obj) {
        if (!(obj instanceof Matrix3D))
            return super.equals(obj);

        Matrix3D m = (Matrix3D) obj;

        return (m.m00 == m00) && (m.m10 == m10) && (m.m20 == m20)
                && (m.m01 == m01) && (m.m11 == m11) && (m.m21 == m21)
                && (m.m02 == m02) && (m.m12 == m12) && (m.m22 == m22)
                && (m.m03 == m03) && (m.m13 == m13) && (m.m23 == m23);
    }

    /**
     * Returns the sx value (m00).
     */
    public double getScaleX() {
        return m00;
    }

    /**
     * Returns the sy value (m11).
     */
    public double getScaleY() {
        return m11;
    }

    /**
     * Returns the sz value (m22).
     */
    public double getScaleZ() {
        return m22;
    }

    /**
     * Returns the tx value (m03).
     */
    public double getTranslateX() {
        return m03;
    }

    /**
     * Returns the ty value (m13).
     */
    public double getTranslateY() {
        return m13;
    }

    /**
     * Returns the tz value (m23).
     */
    public double getTranslateZ() {
        return m23;
    }

    /**
     * Returns a rotate matrix to rotate theta radians over unit vector (nx, ny,
     * nz).
     * 
     * <pre>
     *   [cos(theta)+nx*nx*(1-cos(theta))    nx*ny*(1-cos(theta))-nz*sin(theta)  nx*nz*(1-cos(theta))+ny*sin(theta)  0   ]
     *   [ny*nx*(1-cos(theta)+nz*sin(theta)  cos(theta)+ny*ny*(1-cos(theta))     ny*nz*(1-cos(theta))-nz*sin(theta)  0   ]
     *   [nz*nx*(1-cos(theta)-ny*sin(theta)  nz*ny*(1-cos(theta))+nx*sin(theta)  cos(theta)+nz*nz*(1-cos(theta))     0   ]
     *   [0                                  0                                   0                                   1   ]
     * </pre>
     * 
     * see "M.Pique, Matrix Techniques, Graphics Gems I [GLASSNER, 1990], pp.
     * 465-469"
     */
    public static Matrix3D getRotateInstance(double theta, double nx,
            double ny, double nz) {
        double cos = Math.cos(theta);
        double sin = Math.sin(theta);
        double one_cos = 1 - cos;
        return new Matrix3D(cos + nx * nx * one_cos, ny * nx * one_cos + nz
                * sin, nz * nx * one_cos - ny * sin,

        nx * ny * one_cos - nz * sin, cos + ny * ny * one_cos, nz * ny
                * one_cos + nx * sin,

        nx * nz * one_cos + ny * sin, ny * nz * one_cos - nx * sin, cos + nz
                * nz * one_cos,

        0, 0, 0);
    }

    /**
     * Returns a rotate matrix to rotate omega over the Z-axis, theta over the
     * X-axis, followed phi over the Y-axis.
     */
    public static Matrix3D getRotateInstance(double phi, double theta,
            double omega) {
        Matrix3D m = new Matrix3D();
        if (omega != 0)
            m.rotate(omega, 0, 0, 1);
        if (theta != 0)
            m.rotate(theta, 1, 0, 0);
        if (phi != 0)
            m.rotate(phi, 0, 1, 0);
        return m;
    }

    /**
     * Rotate phi radians over x-axis, where y to z is a positive rotation.
     * Concatenates with matrix:
     * 
     * <pre>
     *  [1               0               0               0]
     *  [0               cos(phi)        -sin(phi)       0]
     *  [0               sin(phi)        cos(phi)        0]
     *  [0               0               0               1]  
     * </pre>
     * 
     * See "Foley, van Dam, Computer Graphics, 1990, pp.213-222"
     */
    public void rotateX(double phi) {
        double cos = Math.cos(phi);
        double sin = Math.sin(phi);
        set( m00,                    m10,                    m20, 
             m01 * cos + m02 * sin,  m11 * cos + m12 * sin,  m21 * cos + m22 * sin, 
            -m01 * sin + m02 * cos, -m11 * sin + m12 * cos, -m21 * sin + m22 * cos, 
             m03,                    m13,                    m23);
    }

    /**
     * Rotate theta radians over y-axis, where z to x is a positive rotation.
     * Concatenates with matrix:
     * 
     * <pre>
     *  [cos(theta)      0               sin(theta)      0]
     *  [0               1               0               0]
     *  [-sin(theta)     0               cos(theta)      0]
     *  [0               0               0               1]
     * </pre>
     * 
     * See "Foley, van Dam, Computer Graphics, 1990, pp.213-222"
     */
    public void rotateY(double theta) {
        double cos = Math.cos(theta);
        double sin = Math.sin(theta);
        set(m00 * cos - m02 * sin, m10 * cos - m12 * sin, m20 * cos - m22 * sin, 
            m01,                   m11,                   m21, 
            m00 * sin + m02 * cos, m10 * sin + m12 * cos, m20 * sin + m22 * cos, 
            m03,                   m13,                   m23);
    }

    /**
     * Rotate omega radians over z-axis, where x to y is a positive rotation.
     * Concatenates with matrix:
     * 
     * <pre>
     *  [cos(omega)      -sin(omega)     0               0]
     *  [sin(omega)      cos(omega)      0               0]
     *  [0               0               1               0]
     *  [0               0               0               1]
     * </pre>
     * 
     * See "Foley, van Dam, Computer Graphics, 1990, pp.213-222"
     */
    public void rotateZ(double omega) {
        double cos = Math.cos(omega);
        double sin = Math.sin(omega);
        set( m00 * cos + m01 * sin,  m10 * cos + m11 * sin,  m20 * cos + m21 * sin, 
            -m00 * sin + m01 * cos, -m10 * sin + m11 * cos, -m20 * sin + m21 * cos, 
             m02,                   m12,                     m22, 
             m03,                   m13,                     m23);
    }

    /**
     * Rotates theta radians over unit vector (nx, ny, nz). Preconcatenates with
     * the matrix:
     * 
     * <pre>
     *   [cos(theta)+nx*nx*(1-cos(theta))    nx*ny*(1-cos(theta))-nz*sin(theta)  nx*nz*(1-cos(theta))+ny*sin(theta)  0   ]
     *   [ny*nx*(1-cos(theta)+nz*sin(theta)  cos(theta)+ny*ny*(1-cos(theta))     ny*nz*(1-cos(theta))-nz*sin(theta)  0   ]
     *   [nz*nx*(1-cos(theta)-ny*sin(theta)  nz*ny*(1-cos(theta))+nx*sin(theta)  cos(theta)+nz*nz*(1-cos(theta))     0   ]
     *   [0                                  0                                   0                                   1   ]
     * </pre>
     * 
     * see "M.Pique, Matrix Techniques, Graphics Gems I [GLASSNER, 1990], pp.
     * 465-469"
     */
    public void rotate(double theta, double nx, double ny, double nz) {
        preConcatenate(Matrix3D.getRotateInstance(theta, nx, ny, nz));
    }

    /**
     * Scales by sx, sy and sz. Preconcatenates with the matrix:
     * 
     * <pre>
     *   [   sx              0               0               0       ]
     *   [   0               sy              0               0       ]
     *   [   0               0               sz              0       ]
     *   [   0               0               0               1       ]
     * </pre>
     */
    public void scale(double sx, double sy, double sz) {
        m00 *= sx;
        m01 *= sx;
        m02 *= sx;
        m03 *= sx;
        m10 *= sy;
        m11 *= sy;
        m12 *= sy;
        m13 *= sy;
        m20 *= sz;
        m21 *= sz;
        m22 *= sz;
        m23 *= sz;
        orthoNormalized = false;
    }

    /**
     * Shears by shx and shy. Preconcatenates with the matrix:
     * 
     * <pre>
     *   [   1               shx             0               0       ]
     *   [   shy             1               0               0       ]
     *   [   0               0               1               0       ]
     *   [   0               0               0               1       ]
     * </pre>
     */
    public void shear(double shx, double shy) {
        set(m00 + shx * m10, m10 + shy * m00, m20,

        m01 + shx * m11, m11 + shy * m01, m21,

        m02 + shx * m12, m12 + shy * m02, m22,

        m03 + shx * m13, m13 + shy * m03, m23);
    }

    /**
     * Translates by tx, ty and tz. Preconcatenates with the matrix:
     * 
     * <pre>
     *   [   1               0               0               tx      ]
     *   [   0               1               0               ty      ]
     *   [   0               0               1               tz      ]
     *   [   0               0               0               1       ]
     * </pre>
     */
    public void translate(double tx, double ty, double tz) {
        m03 += tx;
        m13 += ty;
        m23 += tz;
    }

    /**
     * Model-Translates by tx, ty and tz. Concatenates with the matrix:
     * 
     * <pre>
     *   [   1               0               0               tx      ]
     *   [   0               1               0               ty      ]
     *   [   0               0               1               tz      ]
     *   [   0               0               0               1       ]
     * </pre>
     */
    public void modelTranslate(double tx, double ty, double tz) {
        // m00 m01 m02 m00*tx+m01*ty+m02*tz+m03
        // m10 m11 m12 m10*tx+m11*ty+m12*tz+m13
        // m20 m21 m13 m20*tx+m21*ty+m22*tz+m23
        m03 += m00 * tx + m01 * ty + m02 * tz;
        m13 += m10 * tx + m11 * ty + m12 * tz;
        m23 += m20 * tx + m21 * ty + m22 * tz;
    }

    private void set(double m00, double m10, double m20, double m01,
            double m11, double m21, double m02, double m12, double m22,
            double m03, double m13, double m23) {
        this.m00 = m00;
        this.m10 = m10;
        this.m20 = m20;

        this.m01 = m01;
        this.m11 = m11;
        this.m21 = m21;

        this.m02 = m02;
        this.m12 = m12;
        this.m22 = m22;

        this.m03 = m03;
        this.m13 = m13;
        this.m23 = m23;

        this.orthoNormalized = false;
    }

    /**
     * Returns the angle and unit vector of the rotation matrix of this matrix.
     */
    public double getRotation(double[] n) {
        Matrix3D m = orthoNormalized ? this : createOrthonormalized();
        double cosTheta = (m.m00 + m.m11 + m.m22 - 1) / 2;
        double theta = Math.acos(cosTheta);
        double twoSinTheta = 2 * Math.sin(theta);
        if (twoSinTheta == 0) {
            n[0] = n[1] = n[2] = 0;
        } else {
            n[0] = (m.m21 - m.m12) / twoSinTheta;
            n[1] = (m.m02 - m.m20) / twoSinTheta;
            n[2] = (m.m01 - m.m10) / twoSinTheta;
        }
        return theta;
    }

    /**
     * Returns the determinant.
     */
    public double getDeterminant() {
        return (m00 * m11 - m01 * m10) * m22 - (m00 * m12 - m02 * m10) * m21
                + (m01 * m12 - m02 * m11) * m20;
    }

    /**
     * Create and return the orthonormalized rotation matrix by using the
     * Gram-Schmidt algorithm.
     */
    public Matrix3D createOrthonormalized() {
        double v1length, v2length, v3length;
        double p1, p2;
        Matrix3D r = new Matrix3D();
        r.m00 = m00;
        r.m10 = m10;
        r.m20 = m20;

        v1length = r.m00 * r.m00 + r.m10 * r.m10 + r.m20 * r.m20;
        p1 = (m01 * r.m00 + m11 * r.m10 + m21 * r.m20) / v1length;
        r.m01 = m01 - p1 * m00;
        r.m11 = m11 - p1 * m10;
        r.m21 = m21 - p1 * m20;

        v2length = r.m01 * r.m01 + r.m11 * r.m11 + r.m21 * r.m21;
        p1 = (m02 * r.m00 + m12 * r.m10 + m22 * r.m20) / v1length;
        p2 = (m02 * r.m01 + m12 * r.m11 + m22 * r.m21) / v2length;
        r.m02 = m02 - p1 * m00 - p2 * m01;
        r.m12 = m12 - p1 * m10 - p2 * m11;
        r.m22 = m22 - p1 * m20 - p2 * m21;

        v3length = r.m02 * r.m02 + r.m12 * r.m12 + r.m22 * r.m22;

        double v1n = Math.sqrt(v1length);
        double v2n = Math.sqrt(v2length);
        double v3n = Math.sqrt(v3length);
        r.m00 /= v1n;
        r.m10 /= v1n;
        r.m20 /= v1n;
        r.m01 /= v2n;
        r.m11 /= v2n;
        r.m21 /= v2n;
        r.m02 /= v3n;
        r.m12 /= v3n;
        r.m22 /= v3n;

        return r;
    }

    /**
     * Create and returnes a normalized matrix of this matrix.
     */
    public Matrix3D createNormalized() {
        Matrix3D r = (Matrix3D) clone();

        double v1n = Math.sqrt(r.m00 * r.m00 + r.m10 * r.m10 + r.m20 * r.m20);
        double v2n = Math.sqrt(r.m01 * r.m01 + r.m11 * r.m11 + r.m21 * r.m21);
        double v3n = Math.sqrt(r.m02 * r.m02 + r.m12 * r.m12 + r.m22 * r.m22);
        r.m00 /= v1n;
        r.m10 /= v1n;
        r.m20 /= v1n;
        r.m01 /= v2n;
        r.m11 /= v2n;
        r.m21 /= v2n;
        r.m02 /= v3n;
        r.m12 /= v3n;
        r.m22 /= v3n;

        return r;
    }

    /**
     * Creates and returns the inverse of this matrix.
     */
    public Matrix3D createInverse() throws NoninvertibleTransformException {
        double d = getDeterminant();
        if (Math.abs(d) <= Double.MIN_VALUE)
            throw new NoninvertibleTransformException(
                    "Matrix3D cannot be inverted. Determinant: " + d + "\n"
                            + toString());

        return new Matrix3D(
                (m11 * m22 - m12 * m21) / d,
                (m12 * m20 - m10 * m22) / d,
                (m10 * m21 - m11 * m20) / d,

                (m21 * m02 - m22 * m01) / d,
                (m22 * m00 - m20 * m02) / d,
                (m20 * m01 - m21 * m00) / d,

                (m01 * m12 - m02 * m11) / d,
                (m02 * m10 - m00 * m12) / d,
                (m00 * m11 - m01 * m10) / d,

                (m01 * (m13 * m22 - m12 * m23) + m02 * (m11 * m23 - m13 * m21) + m03
                        * (m12 * m21 - m11 * m22))
                        / d, (m02 * (m13 * m20 - m10 * m23) + m03
                        * (m10 * m22 - m12 * m20) + m00
                        * (m12 * m23 - m13 * m22))
                        / d, (m03 * (m11 * m20 - m10 * m21) + m00
                        * (m13 * m21 - m11 * m23) + m01
                        * (m10 * m23 - m13 * m20))
                        / d);
    }

    /**
     * Concatenates this matrix with m, such that M' = M * m.
     */
    public void concatenate(Matrix3D m) {
        set(m00 * m.m00 + m01 * m.m10 + m02 * m.m20, m10 * m.m00 + m11 * m.m10
                + m12 * m.m20, m20 * m.m00 + m21 * m.m10 + m22 * m.m20,

        m00 * m.m01 + m01 * m.m11 + m02 * m.m21, m10 * m.m01 + m11 * m.m11
                + m12 * m.m21, m20 * m.m01 + m21 * m.m11 + m22 * m.m21,

        m00 * m.m02 + m01 * m.m12 + m02 * m.m22, m10 * m.m02 + m11 * m.m12
                + m12 * m.m22, m20 * m.m02 + m21 * m.m12 + m22 * m.m22,

        m00 * m.m03 + m01 * m.m13 + m02 * m.m23 + m03, m10 * m.m03 + m11
                * m.m13 + m12 * m.m23 + m13, m20 * m.m03 + m21 * m.m13 + m22
                * m.m23 + m23);
    }

    /**
     * Pre-concatenates this matrix with m, such that M' = m * M.
     */
    public void preConcatenate(Matrix3D m) {
        set(m.m00 * m00 + m.m01 * m10 + m.m02 * m20, m.m10 * m00 + m.m11 * m10
                + m.m12 * m20, m.m20 * m00 + m.m21 * m10 + m.m22 * m20,

        m.m00 * m01 + m.m01 * m11 + m.m02 * m21, m.m10 * m01 + m.m11 * m11
                + m.m12 * m21, m.m20 * m01 + m.m21 * m11 + m.m22 * m21,

        m.m00 * m02 + m.m01 * m12 + m.m02 * m22, m.m10 * m02 + m.m11 * m12
                + m.m12 * m22, m.m20 * m02 + m.m21 * m12 + m.m22 * m22,

        m.m00 * m03 + m.m01 * m13 + m.m02 * m23 + m.m03, m.m10 * m03 + m.m11
                * m13 + m.m12 * m23 + m.m13, m.m20 * m03 + m.m21 * m13 + m.m22
                * m23 + m.m23);
    }

    /**
     * Transforms xyz by this matrix, using translation if delta is false. The
     * input array is returned.
     */
    public double[] transform(double[] xyz, boolean delta) {
        int d = delta ? 0 : 1;
        double x = xyz[X];
        double y = xyz[Y];
        double z = xyz[Z];

        xyz[U] = m00 * x + m01 * y + m02 * z + m03 * d;
        xyz[V] = m10 * x + m11 * y + m12 * z + m13 * d;
        xyz[W] = m20 * x + m21 * y + m22 * z + m23 * d;

        return xyz;
    }

    /**
     * Transforms xyz by this matrix for n points, using translation if delta is
     * false. The input array is returned.
     */
    public double[][] transform(double[][] xyz, int n, boolean delta) {
        int d = delta ? 0 : 1;
        for (int i = 0; i < n; i++) {
            double x = xyz[X][i];
            double y = xyz[Y][i];
            double z = xyz[Z][i];
            xyz[U][i] = m00 * x + m01 * y + m02 * z + m03 * d;
            xyz[V][i] = m10 * x + m11 * y + m12 * z + m13 * d;
            xyz[W][i] = m20 * x + m21 * y + m22 * z + m23 * d;
        }
        return xyz;
    }

    public String toString() {
        StringBuffer s = new StringBuffer();
        s.append("Matrix3D: ");
        s.append(fmt.format(m00));
        s.append(", ");
        s.append(fmt.format(m01));
        s.append(", ");
        s.append(fmt.format(m02));
        s.append(", ");
        s.append(fmt.format(m03));
        s.append("\n");

        s.append("          ");
        s.append(fmt.format(m10));
        s.append(", ");
        s.append(fmt.format(m11));
        s.append(", ");
        s.append(fmt.format(m12));
        s.append(", ");
        s.append(fmt.format(m13));
        s.append("\n");

        s.append("          ");
        s.append(fmt.format(m20));
        s.append(", ");
        s.append(fmt.format(m21));
        s.append(", ");
        s.append(fmt.format(m22));
        s.append(", ");
        s.append(fmt.format(m23));
        s.append("\n");

        return s.toString();
    }

    //
    // XMLIO
    //
    public void save(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        nodeEl.setAttribute("m00", Double.toString(m00));
        nodeEl.setAttribute("m10", Double.toString(m10));
        nodeEl.setAttribute("m20", Double.toString(m20));

        nodeEl.setAttribute("m01", Double.toString(m01));
        nodeEl.setAttribute("m11", Double.toString(m11));
        nodeEl.setAttribute("m21", Double.toString(m21));

        nodeEl.setAttribute("m02", Double.toString(m02));
        nodeEl.setAttribute("m12", Double.toString(m12));
        nodeEl.setAttribute("m22", Double.toString(m22));

        nodeEl.setAttribute("m03", Double.toString(m03));
        nodeEl.setAttribute("m13", Double.toString(m13));
        nodeEl.setAttribute("m23", Double.toString(m23));
    }

    public void restore(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        try {
            set(nodeEl.getAttribute("m00").getDoubleValue(), nodeEl
                    .getAttribute("m10").getDoubleValue(), nodeEl.getAttribute(
                    "m20").getDoubleValue(),

            nodeEl.getAttribute("m01").getDoubleValue(), nodeEl.getAttribute(
                    "m11").getDoubleValue(), nodeEl.getAttribute("m21")
                    .getDoubleValue(),

            nodeEl.getAttribute("m02").getDoubleValue(), nodeEl.getAttribute(
                    "m12").getDoubleValue(), nodeEl.getAttribute("m22")
                    .getDoubleValue(),

            nodeEl.getAttribute("m03").getDoubleValue(), nodeEl.getAttribute(
                    "m13").getDoubleValue(), nodeEl.getAttribute("m23")
                    .getDoubleValue());
        } catch (DataConversionException dce) {
            throw new IllegalArgumentException(getClass() + ": "
                    + dce.getMessage());
        }
    }
}
