// Copyright 2005-2006, FreeHEP.
package hep.wired.util.slider;


import hep.wired.util.SwingEventMulticaster;

import java.awt.BorderLayout;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.text.DecimalFormat;
import java.text.Format;
import java.util.Hashtable;

import javax.swing.BoxLayout;
import javax.swing.DefaultBoundedRangeModel;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.plaf.UIResource;

import org.freehep.util.ScientificFormat;

/**
 * FIXME WIRED-440 Should move to freehep-swing ?
 * 
 * @author Mark Donszelmann
 * @version $Id: NumberSlider.java 8598 2006-08-14 20:38:02Z duns $
 */
public class NumberSlider extends JPanel implements ChangeListener, SwingConstants {

    private JSlider slider;
    
    private BoundedIntegerRangeModel intModel;
    private BoundedDoubleRangeModel doubleModel; 

    private Format intFormat = new DecimalFormat();
    private Format doubleFormat = new ScientificFormat(5, 8, false);        

    private static final int bins = 1000000;
    private boolean logarithmic;
    private double factor;
    
    private ChangeListener listener;
    
    private final int majorTicks = 10;

    public NumberSlider() {
        this(VERTICAL, 0, 100, 0);
    }
    
    public NumberSlider(int orientation, double min, double max, double def) {
        this(new DefaultBoundedDoubleRangeModel(def, 0, min, max));
    }
    
    public NumberSlider(int orientation, int min, int max, int def) {
        this(new DefaultBoundedIntegerRangeModel(def, 0, min, max));
    }
    
    public NumberSlider(BoundedNumberRangeModel model) {
        this(VERTICAL, model);
    }
    
    public NumberSlider(int orientation, BoundedNumberRangeModel model) {
        setLayout(new BorderLayout());
        slider = new JSlider(VERTICAL, 0, 100, 0);
        addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent event) {
                slider.setToolTipText(isEnabled() ? getFormat().format(getValue()) : null);
             }
        }); 
        slider.addMouseWheelListener(new MouseWheelListener() {
            public void mouseWheelMoved(MouseWheelEvent e) {
                if (slider.isEnabled()) {
                    int delta = e.getWheelRotation();
                    delta *= (slider.getMaximum() - slider.getMinimum()) / 20;
//                    delta *= slider.getMinorTickSpacing() > 0 ? slider.getMinorTickSpacing() : slider.getMajorTickSpacing();
                    delta = e.isControlDown() ? delta / 5 : delta;
                    delta = e.isShiftDown() ? delta / 2 : delta;
                    slider.setValue(slider.getValue()-delta);
                }
            } 
        });
        add(slider, BorderLayout.CENTER);
                
        setOrientation(orientation);
        setModel(model);
        setPaintTicks(true);
        setPaintLabels(true);
        setPaintTrack(true);
        setLogarithmic(false);
    }
    
    public BoundedNumberRangeModel getModel() {
        if (intModel != null) return intModel;
        return doubleModel;
    }
    
    public void setModel(BoundedNumberRangeModel model) {
        if (intModel != null) intModel.removeChangeListener(this);
        if (doubleModel != null) doubleModel.removeChangeListener(this);
        intModel = null;
        doubleModel = null;
        
        if (model instanceof BoundedIntegerRangeModel) {
            intModel = (BoundedIntegerRangeModel)model;
            intModel.addChangeListener(this);
            slider.setModel(new DefaultBoundedRangeModel() {
                public int getExtent() {
                    return intModel.getExtent();
                }
                
                public int getMinimum() {
                    return intModel.getMinimum();
                }
                
                public int getMaximum() {
                    return intModel.getMaximum();
                }
                
                public int getValue() {
                    return intModel.getValue();
                }
                
                public void setRangeProperties(int value, int extent, int min, int max, boolean adjusting) {
                    super.setRangeProperties(value, extent, min, max, adjusting);
                    intModel.setRangeProperties(value, extent, intModel.getMinimum(), intModel.getMaximum(), intModel.getDefaultValue(), adjusting);
                }
            });
        } else if (model instanceof BoundedDoubleRangeModel) {
            doubleModel = (BoundedDoubleRangeModel)model;
            doubleModel.addChangeListener(this);
            slider.setModel(new DefaultBoundedRangeModel() {
                public int getExtent() {
                    return (int)Math.round(doubleModel.getExtent() / factor);
                }
                
                public int getMinimum() {
                    return 0;
                }
                
                public int getMaximum() {
                    return bins;
                }
                
                public int getValue() {
                    double value = isLogarithmic() ? log10(doubleModel.getValue()) - log10(doubleModel.getMinimum()) : doubleModel.getValue() - doubleModel.getMinimum();
                    int r = (int)Math.round(value / factor);
                    return r;
                }
                
                public void setRangeProperties(int value, int extent, int min, int max, boolean adjusting) {
                    super.setRangeProperties(value, extent, min, max, adjusting);
                    
                    double doubleValue = NumberSlider.this.getValue(value);                     
                    double doubleExtent = extent * factor;
                    doubleModel.setRangeProperties(doubleValue, 
                                                   doubleExtent, 
                                                   doubleModel.getMinimum(), 
                                                   doubleModel.getMaximum(), 
                                                   doubleModel.getDefaultValue(), 
                                                   adjusting);
               }
            });
        } else {
            throw new IllegalArgumentException(getClass()+".setModel(...), model not in class range: "+model.getClass());
        }
        
        stateChanged(new ChangeEvent(this));
    }
    
    public void setEnabled(boolean state) {
        super.setEnabled(state);
        
        slider.setEnabled(state);
        slider.setToolTipText(state ? getFormat().format(getValue()) : null);
    }
    
    public boolean canBeLogarithmic() {
        if (intModel != null) return false;
        return doubleModel.getMinimum() * doubleModel.getMaximum() > 0;
    }

    public boolean isLogarithmic() {
        return logarithmic;
    }

    public void setLogarithmic(boolean state) {
        if (state == logarithmic) return;
        if (state && !canBeLogarithmic()) {
            throw new IllegalArgumentException(
                    "min and max need to be on the same side of zero for a log slider and cannot be zero ("
                            + getMinimum()
                            + ", "
                            + getMaximum() + ")");
        } 
        logarithmic = state;
        stateChanged(new ChangeEvent(this));
    }

    public Format getFormat() {
        return (intModel != null) ? intFormat : doubleFormat;
    }

    public void setFormat(Format format) {
        if (intModel != null) {
            intFormat = format;
        } else {
            doubleFormat = format;
        }
        stateChanged(new ChangeEvent(this));
    }

    public int getOrientation() {
        return slider.getOrientation();
    }
    
    public void setOrientation(int orientation) {
        slider.setOrientation(orientation);
    }
    
    public Number getValue() {
        if (intModel != null) return new Integer(intModel.getValue());
        return new Double(doubleModel.getValue());
    }
    
    public Number getMinimum() {
        if (intModel != null) return new Integer(intModel.getMinimum());
        return new Double(doubleModel.getMinimum());
    }
 
    public Number getMaximum() {
        if (intModel != null) return new Integer(intModel.getMaximum());
        return new Double(doubleModel.getMaximum());
    }
 
    // forward events from model
    public void stateChanged(ChangeEvent e) {
        int diff = 0;
        int increment = 0;
        if (intModel != null) {
            factor = isLogarithmic() ? (log10(intModel.getMaximum()) - log10(intModel.getMinimum())) / bins : 1.0;
            slider.getModel().setRangeProperties(intModel.getValue(), 
                                                 intModel.getExtent(), 
                                                 intModel.getMinimum(),
                                                 intModel.getMaximum(), 
                                                 intModel.getValueIsAdjusting());
            diff = (int)Math.round((slider.getMaximum() - slider.getMinimum()));
            increment = diff >= majorTicks ? diff / majorTicks : 1;
            slider.setLabelTable(slider.createStandardLabels(increment));
        } else if (doubleModel != null) {
            double range = isLogarithmic() ? log10(doubleModel.getMaximum()) - log10(doubleModel.getMinimum()) 
                                           : doubleModel.getMaximum() - doubleModel.getMinimum();
            factor = range / bins;
            
            int intValue = isLogarithmic() ? (int)Math.round(((log10(doubleModel.getValue()) - log10(doubleModel.getMinimum())) / factor))
                                           : (int)Math.round((doubleModel.getValue() - doubleModel.getMinimum()) / factor);
            int intExtent = (int)Math.round(doubleModel.getExtent() / factor);
            
            slider.getModel().setRangeProperties(intValue, 
                                                 intExtent, 
                                                 0,
                                                 bins, 
                                                 doubleModel.getValueIsAdjusting());
            int ticks = isLogarithmic() ? (int)Math.round(log10(doubleModel.getMaximum())) - (int)Math.round(log10(doubleModel.getMinimum())) 
                                        : majorTicks;
            diff = (int)Math.round((slider.getMaximum() - slider.getMinimum()));
            increment = diff >= ticks ? diff / ticks : 1;
            slider.setLabelTable(createStandardLabels(increment));
        }
        slider.setMajorTickSpacing(increment);
        if (isLogarithmic()) {
            slider.setMinorTickSpacing(0);
        } else {
            slider.setMinorTickSpacing(increment / 2);
        }
        
        fireStateChanged(e);
    }

    public void addChangeListener(ChangeListener a) {
        listener = SwingEventMulticaster.add(listener, a);
    }
    
    public void removeChangeListener(ChangeListener r) {
        listener = SwingEventMulticaster.remove(listener, r);
    }
    
    private void fireStateChanged(ChangeEvent e) {
        if (listener != null) {
            listener.stateChanged(e);
        }
    }
    
    public void setPaintTicks(boolean state) {
        slider.setPaintTicks(state);
    }
    
    public boolean getPaintTicks() {
        return slider.getPaintTicks();
    }
    
    public void setPaintLabels(boolean state) {
        slider.setPaintLabels(state);
    }
    
    public boolean getPaintLabels() {
        return slider.getPaintLabels();
    }
    
    public void setPaintTrack(boolean state) {
        slider.setPaintTrack(state); 
    }
    
    public boolean getPaintTrack() {
        return slider.getPaintTrack();
    }

    private static double log10(double x) {
        return Math.log(x) / Math.log(10);
    }

    private static double pow10(double x) {
        return Math.pow(10, x);
    }
    
    // only for doubleModel
    private Hashtable createStandardLabels(int increment) {
        if (increment <= 0) {
            throw new IllegalArgumentException(
                    "Label incremement must be > 0: " + increment);
        }

        class LabelUIResource extends JLabel implements UIResource {
            public LabelUIResource(String text) {
                super(text, JLabel.CENTER);
                setName("Slider.label");
            }
        }

        Hashtable table = new Hashtable();
       
        int labelStart = slider.getMinimum();
        int labelMaximum = slider.getMaximum();
        for (int labelIndex = labelStart; labelIndex <= labelMaximum; labelIndex += increment) {
            Double value = new Double(getValue(labelIndex - labelStart));
            String field = getFormat().format(value);
            table.put(new Integer(labelIndex), new LabelUIResource(field));
        }

        return table;
    }
    
    private double getValue(int bin) {
        return isLogarithmic() ? pow10(bin * factor + log10(doubleModel.getMinimum())) 
                               : (bin * factor) + doubleModel.getMinimum(); 
    }
    
    private static void add(JPanel p, final NumberSlider s) {
        s.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent e) {
                System.err.println(e.getSource().hashCode() + " "
                        + s.getValue());
            }
        });
        if (s.canBeLogarithmic()) {
            JPanel t = new JPanel();
            p.add(t);

            t.setLayout(new BorderLayout());
            t.add(s, BorderLayout.CENTER);
            
            final JCheckBox log = new JCheckBox("log", s.isLogarithmic());
            t.add(log, BorderLayout.SOUTH);
            
            log.addChangeListener(new ChangeListener() {
                public void stateChanged(ChangeEvent e) {
                    s.setLogarithmic(log.isSelected());
                }
            });
        } else {
            p.add(s);
        }
    }

    public static void main(String[] args) {
        JPanel p = new JPanel();
        p.setLayout(new BoxLayout(p, BoxLayout.X_AXIS));

/*        add(p, new NumberSlider(VERTICAL, -40, 360, 0));
        add(p, new NumberSlider(VERTICAL, 0, 2000, 100));
        add(p, new NumberSlider(VERTICAL, -100, 200, -30));
        add(p, new NumberSlider(VERTICAL, -100.0, 200.0, -30.0));
        add(p, new NumberSlider(VERTICAL, -0.01, 0.03, 0));
        add(p, new NumberSlider(VERTICAL, 0, 0.00001, 0));
        add(p, new NumberSlider(VERTICAL, 0.01, 1000, 100));
*/
        NumberSlider n = new NumberSlider(VERTICAL, 0.01, 1000, 100);
        n.setLogarithmic(true);
        add(p, n);
        
        JFrame f = new JFrame();
        f.getContentPane().add(p);
        f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        f.setSize(800, 600);
        f.setVisible(true);
    }

}
