package hep.wired.util.tree;

import java.util.*;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreeSelectionModel;

import javax.swing.tree.TreePath;

/**
 * 
 * @author Dmitry Onoprienko
 * @version $Id: $
 */
public class WiredTreeModel extends DefaultTreeModel implements TreeExpansionListener {

//  -- Private parts : ---------------------------------------------------------

  protected WiredTreeSelectionModel _selection;
  private TreePath _rootPath;

// -- Construction : -----------------------------------------------------------

  public WiredTreeModel(WiredTreeNode root) {
    this(root, false);
  }

  public WiredTreeModel(WiredTreeNode root, boolean asksAllowsChildren) {
    super(root, asksAllowsChildren);
    _selection = new WiredTreeSelectionModel();
    _selection.setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
    _rootPath = (root == null) ? null : new TreePath(root);
    if (root != null) root.setExpanded(true);
  }

// -- Setters : ----------------------------------------------------------------

  public void setRoot(TreeNode root) {
    _rootPath = (root == null) ? null : new TreePath(root);
    if (root != null) ((WiredTreeNode)root).setExpanded(true);
    super.setRoot(root);
  }

// -- Getters : ----------------------------------------------------------------

  public WiredTreeSelectionModel getSelectionModel() {
    return _selection;
  }

  public TreePath getRootPath() {
    return _rootPath;
  }

  public WiredTreeNode getRoot() {
    return (WiredTreeNode) root;
  }

  public Set<? extends Object> getMarkedObjects() {
    final Set<Object> out = new HashSet();
    TreeVisitor v = new TreeVisitor() {
      public boolean visit(WiredTreeNode node) {
        if (node.isMarkedNode() && ! node.isHidden()) out.add(node.getPayload());
        return node.isMarkedBranch();
      }
    };
    v.visitNodes(this);
    return out;
  }

// -- Handling mark state listeners : ------------------------------------------

  public interface MarkListener extends EventListener {
    public void markChanged(MarkEvent e);
  }

  public static class MarkEvent  {
    protected WiredTreeModel _source;
    protected WiredTreeNode _node;
    public MarkEvent(WiredTreeModel source) {_source = source;}
    public MarkEvent(WiredTreeModel source, WiredTreeNode node) {_source = source; _node = node;}
    public WiredTreeModel getSource() {return _source;}
    public WiredTreeNode getNode() {return _node == null ? (WiredTreeNode)_source.getRoot() : _node;}
    public String toString() {return getClass().getName() +" on "+ this;}
  }

  public void addMarkListener(MarkListener l) {
    listenerList.add(MarkListener.class, l);
  }

  public void removeMarkListener(MarkListener l) {
    listenerList.remove(MarkListener.class, l);
   }

  public void fireMarkChanged() {
    fireMarkChanged(null);
  }

  public void fireMarkChanged(WiredTreeNode node) {
    Object[] listeners = listenerList.getListenerList();
    MarkEvent e = null;
    for (int i = listeners.length - 2; i >= 0; i -= 2) {
      if (listeners[i] == MarkListener.class) {
        if (e == null) e = new MarkEvent(this, node);
        ((MarkListener) listeners[i + 1]).markChanged(e);
      }
    }
  }

// -- Keep track of which nodes are expanded (this info is private in JTree for sun-knows-what reason) : -----

  public void treeCollapsed(TreeExpansionEvent event) {
    //System.out.println("-- "+ event.getPath());
    ((WiredTreeNode)event.getPath().getLastPathComponent()).removeFlag(WiredTreeNode.Flag.EXPANDED);
  }

  public void treeExpanded(TreeExpansionEvent event) {
    //System.out.println("++ "+ event.getPath());
    ((WiredTreeNode)event.getPath().getLastPathComponent()).addFlag(WiredTreeNode.Flag.EXPANDED);
  }

// -- Visitor class for tree nodes : -------------------------------------------

  /**
   * Convenience class that simplifies visiting nodes of a WiredTree.
   * Override <tt>visit(WiredTreeNode node)</tt> or <tt>visit(TreePath path)</tt> to do
   * something useful, then call <tt>visitNodes</tt> or <tt>visitPaths</tt>, respectively.
   * When overriding <tt>visit</tt> method, return <tt>true</tt> if you are interested in visiting
   * the node's descendants.
   * <p>
   * Range of tree levels to be visited can be limited by giving <tt>minLevel</tt> and
   * <tt>maxLevel</tt> arguments to <tt>visitNodes</tt> and <tt>visitPaths</tt>.
   * Levels are numbered from 0, starting with the tree root or the node specified in a call.
   * Levels <tt>minLevel</tt> through <tt>maxLevel-1</tt> will be visited (negative
   * <tt>maxLevel</tt> means no limit).
   */
  public static abstract class TreeVisitor {
    public int _minLevel, _maxLevel;
    public boolean visit(WiredTreeNode node) {return true;}
    public boolean visit(TreePath path) {return true;}
    public void visitNodes(WiredTreeNode root) {visitNodes(root, 0, Integer.MAX_VALUE);}
    public void visitNodes(WiredTreeNode root, int maxLevel) {visitNodes(root, 0, maxLevel);}
    public void visitNodes(WiredTreeNode root, int minLevel, int maxLevel) {
      if (root != null) {
        _minLevel = minLevel;
        _maxLevel = maxLevel < 0 ? Integer.MAX_VALUE : maxLevel;
        visitTreeNodes(root,0);
      }
    }
    public void visitNodes(WiredTree tree) {if (tree != null) visitNodes(tree.getModel());}
    public void visitNodes(WiredTree tree, int maxLevel) {if (tree != null) visitNodes(tree.getModel(), maxLevel);}
    public void visitNodes(WiredTree tree, int minLevel, int maxLevel) {if (tree != null) visitNodes(tree.getModel(), minLevel, maxLevel);}
    public void visitNodes(WiredTreeModel model) {if (model != null) visitNodes((WiredTreeNode) model.getRoot());}
    public void visitNodes(WiredTreeModel model, int maxLevel) {if (model != null) visitNodes((WiredTreeNode) model.getRoot(), maxLevel);}
    public void visitNodes(WiredTreeModel model, int minLevel, int maxLevel) {if (model != null) visitNodes((WiredTreeNode) model.getRoot(), minLevel, maxLevel);}
    private void visitTreeNodes(WiredTreeNode root, int level) {
      if ((level < _minLevel || visit(root)) && level < _maxLevel) {
        for (Enumeration<WiredTreeNode> e = root.children(); e.hasMoreElements();) {
          visitTreeNodes(e.nextElement(), level+1);
        }
      }
    }
    public void visitPaths(TreePath root) {visitPaths(root, 0, Integer.MAX_VALUE);}
    public void visitPaths(TreePath root, int maxLevel) {visitPaths(root, 0, maxLevel);}
    public void visitPaths(TreePath root, int minLevel, int maxLevel) {
      if (root != null) {
        _minLevel = minLevel;
        _maxLevel = maxLevel < 0 ? Integer.MAX_VALUE : maxLevel;
        visitTreePaths(root,0);
      }
    }
    public void visitPaths(WiredTree tree) {if (tree != null) visitPaths(tree.getModel());}
    public void visitPaths(WiredTree tree, int maxLevel) {if (tree != null) visitPaths(tree.getModel(), maxLevel);}
    public void visitPaths(WiredTree tree, int minLevel, int maxLevel) {if (tree != null) visitPaths(tree.getModel(), minLevel, maxLevel);}
    public void visitPaths(WiredTreeModel model) {if (model != null) visitPaths(model.getRootPath());}
    public void visitPaths(WiredTreeModel model, int maxLevel) {if (model != null) visitPaths(model.getRootPath(), maxLevel);}
    public void visitPaths(WiredTreeModel model, int minLevel, int maxLevel) {if (model != null) visitPaths(model.getRootPath(), minLevel, maxLevel);}
    private void visitTreePaths(TreePath root, int level) {
      if ((level < _minLevel || visit(root)) && level < _maxLevel) {
        for (Enumeration<WiredTreeNode> e = ((WiredTreeNode)root.getLastPathComponent()).children(); e.hasMoreElements();) {
          visitTreePaths(root.pathByAddingChild(e.nextElement()), level+1);
        }
      }
    }
  }

// -----------------------------------------------------------------------------

}
