package hep.wired.util.tree;

import java.util.*;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;

import hep.wired.util.tree.BranchCheckBox.State;
import static  hep.wired.util.tree.WiredTreeNode.Flag.*;

/**
 * 
 * @author Dmitry Onoprienko
 * @version $Id: $
 */
public class WiredTreeNode extends DefaultMutableTreeNode {

  static public enum Flag {MARKED_NODE, MARKED_BRANCH, HAS_UNMARKED_DESCENDANTS, HAS_UNMARKED_DESCENDANTS_VALID,
                                  SELECTED, HAS_SELECTED_DESCENDANTS, HAS_SELECTED_DESCENDANTS_VALID,
                                  HIDDEN,
                                  EXPANDED}

// -- Private parts : ----------------------------------------------------------

  protected EnumSet<Flag> _flag;
  protected Object _payload;

// -- Construction : -----------------------------------------------------------

  public WiredTreeNode(Object payload, Flag... flags) {
    this(payload, (flags.length == 0) ? EnumSet.of(MARKED_NODE, MARKED_BRANCH) : EnumSet.of(flags[0], flags) );
  }

  public WiredTreeNode(Object payload, EnumSet<Flag> flags) {
    super(null);
    _flag = flags;
    _payload = payload;
    super.setUserObject(this);
  }

// -- Setters : ----------------------------------------------------------------

  public void setPayload(Object payload) {
    _payload = payload;
  }

  public void setFlags(Flag... flags) {
    if (flags.length == 0) {
      _flag = EnumSet.noneOf(Flag.class);
    } else {
      _flag = EnumSet.of(flags[0], flags);
    }
  }

  public void setFlags(EnumSet<Flag> flags) {
    _flag = EnumSet.copyOf(flags);
  }

  public boolean addFlag(Flag flag) {
    return _flag.add(flag);
  }

  public boolean removeFlag(Flag flag) {
    return _flag.remove(flag);
  }

  public boolean setFlag(Flag flag, boolean value) {
    return value ? _flag.add(flag) : _flag.remove(flag);
  }

  public void setMarkedNode(boolean isMarkedNode) {
    if (isMarkedNode) {_flag.add(MARKED_NODE);} else {_flag.remove(MARKED_NODE);}
    WiredTreeNode p = (WiredTreeNode) getParent();
    if (p != null)  p.invalidateHasUnmarkedDescendants();
  }

  public void setMarkedBranch(boolean isMarkedBranch) {
    if (isMarkedBranch) {_flag.add(MARKED_BRANCH);} else {_flag.remove(MARKED_BRANCH);}
    WiredTreeNode p = (WiredTreeNode) getParent();
    if (p != null && p._flag.contains(HAS_UNMARKED_DESCENDANTS) == isMarkedBranch)  {
      p.invalidateHasUnmarkedDescendants();
    }
  }

  public void setSelected(boolean selected) {
    if (selected) {_flag.add(SELECTED);} else {_flag.remove(SELECTED);}
    WiredTreeNode p = (WiredTreeNode) getParent();
    if (p != null && p._flag.contains(HAS_SELECTED_DESCENDANTS) != selected) {
      p.invalidateHasSelectedDescendants();
    }
  }

  public void setHidden(boolean hidden) {
    if (hidden) {_flag.add(HIDDEN);} else {_flag.remove(HIDDEN);}
  }

  public void setExpanded(boolean expanded) {
    if (expanded) {_flag.add(EXPANDED);} else {_flag.remove(EXPANDED);}
  }

  public void add(MutableTreeNode newChild) {
    super.add(newChild);
    invalidateHasUnmarkedDescendants();
  }

// -- Getters : ----------------------------------------------------------------

  public Object getPayload() {
    return _payload;
  }

  public EnumSet<Flag> getFlags() {
    return _flag;
  }

  public boolean isMarkedNode() {
    return _flag.contains(MARKED_NODE);
  }

  public boolean isMarkedBranch() {
    return _flag.contains(MARKED_BRANCH);
  }

  public BranchCheckBox.State getBranchState() {
    if (_flag.add(HAS_UNMARKED_DESCENDANTS_VALID)) {
      _flag.remove(HAS_UNMARKED_DESCENDANTS);
      for (Enumeration<WiredTreeNode> e = children(); e.hasMoreElements();) {
        WiredTreeNode n = e.nextElement();
        if (!n.isMarkedNode() || n.getBranchState() != State.ALL) {
          _flag.add(HAS_UNMARKED_DESCENDANTS);
          break;
        }
      }
    }
    return _flag.contains(MARKED_BRANCH) ? (_flag.contains(HAS_UNMARKED_DESCENDANTS) ? State.PART : State.ALL) : State.NONE;
  }

  public boolean isSelected() {
    return _flag.contains(SELECTED);
  }

  public boolean isHidden() {
    return _flag.contains(HIDDEN);
  }

  public boolean isExpanded() {
    return _flag.contains(EXPANDED);
  }

  public boolean hasSelectedDecendents() {
    if (_flag.add(HAS_SELECTED_DESCENDANTS_VALID)) {
      _flag.remove(HAS_SELECTED_DESCENDANTS);
      for (Enumeration<WiredTreeNode> e = children(); e.hasMoreElements();) {
        WiredTreeNode n = e.nextElement();
        if (n.isSelected() || n.hasSelectedDecendents()) {
          _flag.add(HAS_SELECTED_DESCENDANTS);
          break;
        }
      }
    }
    return _flag.contains(HAS_SELECTED_DESCENDANTS);
  }

  public boolean isInUnmarkedBranch() {
    TreeNode p = getParent();
    while (p != null) {
      if (! ((WiredTreeNode)p).isMarkedBranch()) return true;
      p = p.getParent();
    }
    return false;
  }

  public String toString() {
    return _payload.toString();
  }

// -- Local methods : ----------------------------------------------------------

  protected void invalidateHasUnmarkedDescendants() {
    if (_flag.remove(HAS_UNMARKED_DESCENDANTS_VALID)) {
      TreeNode mother = getParent();
      if (parent != null) ((WiredTreeNode)mother).invalidateHasUnmarkedDescendants();
    }
  }

  protected void invalidateHasSelectedDescendants() {
    if (_flag.remove(HAS_SELECTED_DESCENDANTS_VALID)) {
      TreeNode mother = getParent();
      if (parent != null) ((WiredTreeNode)mother).invalidateHasSelectedDescendants();
    }
  }

// -----------------------------------------------------------------------------

}
