// Copyright 2005, FreeHEP.
package hep.wired.variable;

import java.lang.reflect.Field;

import org.freehep.xml.io.XMLIOManager;
import org.jdom.DataConversionException;
import org.jdom.Element;

/**
 * 
 * @author Mark Donszelmann
 * @version $Id: DoubleVariable.java 8598 2006-08-14 20:38:02Z duns $
 */

public class DoubleVariable extends NumberVariable {

    private double min, max, def;

    public DoubleVariable(String name, Object object, double min, double max,
            double def, String unit, String description) {
        super(name, object, unit, description);
        if (min > max)
            throw new IllegalArgumentException("Variable: min (" + min
                    + ") > max (" + max + ").");
        if ((def > max) || (def < min))
            throw new IllegalArgumentException("Variable: def (" + def
                    + ") not in interval [min,max] [" + min + "," + max + "].");
        this.min = min;
        this.max = max;
        this.def = def;
        setVariable(def);
    }

    /**
     * Returns the value of a named variable.
     */
    public double getDoubleVariable() {
        try {
            Field f = getObject().getClass().getDeclaredField(getName());
            f.setAccessible(true);
            if (f.getType() == double.class) {
                return f.getDouble(getObject());
            } else if (f.getType() == int.class) {
                return f.getInt(getObject());
            }
            throw new IllegalArgumentException("Variable with name '"
                    + getName() + "' not of type 'double' or 'int'.");

        } catch (NoSuchFieldException e) {
            throw new IllegalArgumentException("Variable with name '"
                    + getName() + "' does not exist as 'public' in class '"
                    + getObject().getClass() + "'.");
        } catch (IllegalAccessException e) {
            throw new IllegalArgumentException(
                    "No access to variable with name '" + getName() + "'.");
        }
    }

    public boolean setVariable(double value) {
        value = Math.max(value, getMinimum());
        value = Math.min(value, getMaximum());

        boolean fire = setValue(value);
        if (fire)
            fireStateChanged(VariableChangeEvent.VALUE);
        return fire;
    }

    private boolean setValue(double value) {
        try {
            Field f = getObject().getClass().getDeclaredField(getName());
            f.setAccessible(true);
            if (f.getType() != double.class) {
                throw new IllegalArgumentException("Variable with name '"
                        + getName() + "' not of type 'double'.");
            }

            if (f.getDouble(getObject()) == value)
                return false;

            f.setDouble(getObject(), value);
        } catch (NoSuchFieldException e) {
            throw new IllegalArgumentException("Variable with name '"
                    + getName() + "' does not exist as 'public' in class '"
                    + getObject().getClass() + "'.");
        } catch (IllegalAccessException e) {
            throw new IllegalArgumentException(
                    "No access to variable with name '" + getName() + "'.");
        }
        return true;
    }

    public Object getValue() {
        return new Double(getDoubleVariable());
    }

    public boolean setValue(Object value) {
        if (!(value instanceof Double)) {
            throw new IllegalArgumentException("Variable with name '"
                    + getName() + "' cannot handle value of class '"
                    + value.getClass() + "'.");
        }

        return setVariable(((Double) value).doubleValue());
    }

    public double getMinimum() {
        return min;
    }

    public void setMinimum(double min) {
        if (this.min == min)
            return;

        min = Math.min(min, max);
        this.min = min;
        int type = VariableChangeEvent.MIN;
        if (setValue(Math.max(min, getDoubleVariable())))
            type |= VariableChangeEvent.VALUE;
        fireStateChanged(type);
    }

    public double getMaximum() {
        return max;
    }

    public void setMaximum(double max) {
        if (this.max == max)
            return;

        max = Math.max(max, min);
        this.max = max;
        int type = VariableChangeEvent.MAX;
        if (setValue(Math.min(max, getDoubleVariable())))
            type |= VariableChangeEvent.VALUE;
        fireStateChanged(type);
    }

    public double getDefault() {
        return def;
    }

    public void saveValue(Element element) {
        element.setAttribute("val", Double.toString(getDoubleVariable()));
    }

    public void restoreValue(Element element) {
        try {
            setVariable(element.getAttribute("val").getDoubleValue());
        } catch (DataConversionException dce) {
            throw new IllegalArgumentException(getClass() + ": "
                    + dce.getMessage());
        }
    }

    public String toString() {
        return Double.toString(getDoubleVariable());
    }

    public void save(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {

        nodeEl.setAttribute("min", Double.toString(min));
        nodeEl.setAttribute("max", Double.toString(max));
        nodeEl.setAttribute("def", Double.toString(def));
    }

    public void restore(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        try {
            min = nodeEl.getAttribute("min").getDoubleValue();
            max = nodeEl.getAttribute("max").getDoubleValue();
            def = nodeEl.getAttribute("def").getDoubleValue();
        } catch (DataConversionException dce) {
            throw new IllegalArgumentException(getClass() + ": "
                    + dce.getMessage());
        }
    }
}
