// Copyright 2004, FreeHEP.
package hep.wired.viewport;

import hep.wired.services.ViewPort;

import java.awt.geom.Rectangle2D;

import org.freehep.xml.io.XMLIOManager;
import org.jdom.DataConversionException;

/**
 * Defines a Rectangular Viewport using x0, y0, width and height.
 * 
 * Applies the viewport functions:
 * 
 * <PRE>
 * 
 * u = (x + 1) * width / 2 + x0; v = height - (y + 1) * height / 2 + y0; w = (z +
 * 1) * width / 2 + x0; // NOTE: this is done to be able to calculate bounding
 * boxes in Z.
 * 
 * </PRE>
 * 
 * given rectangle(x0, y0, width, height).
 * 
 * This projects a viewport of ([-1,1],[-1,1]) onto
 * ([x0,y0+height],[x0+width,y0]).
 * 
 * @author Mark Donszelmann
 * @version $Id: RectangularViewPort.java 8598 2006-08-14 20:38:02Z duns $
 */

public class RectangularViewPort extends Rectangle2D.Double implements ViewPort {

    private int componentWidth, componentHeight;

    /**
     * Creates a rectangular viewport of (0, 0, 800, 600).
     */
    public RectangularViewPort() {
        this(0, 0, 800, 600, 800, 600);
    }

    /**
     * Creates a rectangular viewport from given rectangle.
     */
    public RectangularViewPort(Rectangle2D r, int componentWidth,
            int componentHeight) {
        super(r.getX(), r.getY(), r.getWidth(), r.getWidth());
        init(componentWidth, componentHeight);
    }

    /**
     * Creates a rectangular viewport from given parameters.
     */
    public RectangularViewPort(double x, double y, double width, double height,
            int componentWidth, int componentHeight) {
        super(x, y, width, height);
        init(componentWidth, componentHeight);
    }

    private void init(int componentWidth, int componentHeight) {
        this.componentWidth = componentWidth;
        this.componentHeight = componentHeight;
    }

    public int getComponentWidth() {
        return componentWidth;
    }

    public int getComponentHeight() {
        return componentHeight;
    }

    public String getID() {
        return getName();
    }

    public String getName() {
        return "RectangularViewPort";
    }

    public ViewPort copy() {
        return new RectangularViewPort(this, getComponentWidth(),
                getComponentHeight());
    }

    public double[] transform(double[] xyz) {
        xyz[U] = (xyz[X] + 1) * getWidth() / 2.0 + getX();
        xyz[V] = getHeight() - (xyz[Y] + 1) * getHeight() / 2.0 + getY();
        xyz[W] = (xyz[Z] + 1) * getWidth() / 2.0 + getX();
        return xyz;
    }

    public double[][] transform(double[][] xyz, int n) {
        for (int i = 0; i < n; i++) {
            xyz[U][i] = (xyz[X][i] + 1) * getWidth() / 2.0 + getX();
            xyz[V][i] = getHeight() - (xyz[Y][i] + 1) * getHeight() / 2.0
                    + getY();
            xyz[W][i] = (xyz[Z][i] + 1) * getWidth() / 2.0 + getX();
        }
        return xyz;
    }

    public double[] deltaTransform(double[] xyz) {
        xyz[U] *= getWidth() / 2.0;
        xyz[V] *= getHeight() / 2.0;
        xyz[W] *= getWidth() / 2.0;
        return xyz;
    }

    public double[] inverseTransform(double[] uvw) {
        uvw[X] = 2 * (uvw[U] - getX()) / getWidth() - 1;
        uvw[Y] = 2 * (getY() - uvw[V]) / getHeight() + 1;
        uvw[Z] = 2 * (uvw[W] - getX()) / getWidth() - 1;
        return uvw;
    }

    public double[] inverseDeltaTransform(double[] uvw) {
        uvw[X] = 2 * uvw[U] / getWidth();
        uvw[Y] = 2 * uvw[V] / getHeight();
        uvw[Z] = 2 * uvw[W] / getWidth();
        return uvw;
    }

    public String toString() {
        return getClass() + ": " + super.toString();
    }

    //
    // XMLIO
    //
    public void save(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        nodeEl.setAttribute("x0", java.lang.Double.toString(getX()));
        nodeEl.setAttribute("y0", java.lang.Double.toString(getY()));
        nodeEl.setAttribute("width", java.lang.Double.toString(getWidth()));
        nodeEl.setAttribute("height", java.lang.Double.toString(getHeight()));
        nodeEl.setAttribute("componentWidth", Integer
                .toString(getComponentWidth()));
        nodeEl.setAttribute("componentHeight", Integer
                .toString(getComponentHeight()));
    }

    public void restore(XMLIOManager xmlioManager, org.jdom.Element nodeEl) {
        try {
            setRect(nodeEl.getAttribute("x0").getDoubleValue(), nodeEl
                    .getAttribute("y0").getDoubleValue(), nodeEl.getAttribute(
                    "width").getDoubleValue(), nodeEl.getAttribute("height")
                    .getDoubleValue());
            componentWidth = nodeEl.getAttribute("componentWidth")
                    .getIntValue();
            componentHeight = nodeEl.getAttribute("componentHeight")
                    .getIntValue();
        } catch (DataConversionException dce) {
            throw new IllegalArgumentException(getClass() + ": "
                    + dce.getMessage());
        }
    }
}
